# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Schemes for execute chains of jobs in pipelines."""

import abc
import yaml
from functools import partial
from collections.abc import Mapping
import copy

from .logger import logger, log_to_list
from .py_utils import IndexedDict, check_iterable


# TODO: __str__, __repr__, print functions for all classes
# TODO: More logging and debuging


__all__ = ['Config', 'Product', 'Instrument', 'Manager', 'Factory',
           'Stage', 'FrozenError']


def info_dumper(infos):
    """Dump a dictionary information to a formated string.

    Now, it's just a wrapper to yaml.dump, put here to customize if needed.
    """
    return yaml.dump(infos)


class FrozenError(RuntimeError):
    """Pipeline is frozen."""

    pass


class _Frozable():
    _frozen = False

    def __init__(self):
        self._frozen = False

    def __setattr__(self, name, value):
        if self._frozen and name not in self._mutable:
            raise FrozenError('Trying to change an attr while frozen.')
        return super().__setattr__(name, value)

    def __delattr__(self, name):
        if self._frozen:
            raise FrozenError('Trying to delete an attr while frozen.')
        return super().__delattr__(name)

    def freeze(self):
        self._frozen = True
        for v in self.__dict__.values():
            if isinstance(v, _Frozable):
                v.freeze()
        return self

    def unfreeze(self):
        self._frozen = False
        for v in self.__dict__.values():
            if isinstance(v, _Frozable):
                v.unfreeze()
        return self

    @property
    def frozen(self):
        return self._frozen


class Config(dict, _Frozable):
    """Class for generic sotring configs. Like a powered dict."""

    _mutable = ['_frozen', 'logger']

    def __init__(self, *args, **kwargs):
        self._frozen = False
        super().__init__()

        if len(args) > 1:
            raise ValueError('Just one positional argument supported.')
        elif len(args) == 1:
            up = args[0]
            if isinstance(up, Mapping):
                self.update(up)
            elif check_iterable(up):
                for i, v in up:
                    self.__setitem__(i, v)
            else:
                raise TypeError('Argument not supported for dict'
                                ' initialization.')

        for i, v in kwargs.items():
            self.__setitem__(i, v)

    def __setitem__(self, name, value):
        if self._frozen:
            raise FrozenError(f'Tried to change `{name}` with value '
                              '`{value}` while'
                              ' {self.__class__.__name__} is frozen.')

        if isinstance(value, Mapping):
            # Convert to this class if not
            value = self.__class__(value)

        super().__setitem__(name, value)

    def __delitem__(self, name):
        if self._frozen:
            raise FrozenError(f'Tried to delete `{name}` while'
                              ' {self.__class__.__name__} is frozen.')
        return super().__delitem__(name)

    def update(self, dictlike):
        for k, v in dictlike.items():
            self.__setitem__(k, v)

    def freeze(self):
        super().freeze()
        for v in self.values():
            if isinstance(v, _Frozable):
                v.freeze()
        return self

    def unfreeze(self):
        super().unfreeze()
        for v in self.values():
            if isinstance(v, _Frozable):
                v.unfreeze()
        return self


class Instrument(abc.ABC, _Frozable):
    """Store all the informations and needed functions of a instrument."""

    _mutable = ['_frozen']
    _identifier = ''

    def __init__(self, *args, **kwargs):
        self._frozen = False
        super().__init__(*args, **kwargs)

    def list_functions(self):
        """List the class functions."""
        # Pass any callable object that do not start with '_' may cause problem
        # This may pass unwanted functions. Commented out.
        # l = [i.__name__ for i in self.__class__.__dict__.values()
        #      if callable(i) and i.__name__[0] != '_']

        funcs = [i.__name__ for i in self.__class__.__dict__.values()
                 if type(i) in ['function', 'builtin_function_or_method'] and
                 i.__name__[0] != '_']

        # If needed to remove another class function, put here
        for i in ['list_functions']:
            if i in funcs:
                funcs.remove(i)
        return funcs


class Product():
    """Store all informations and data of a product."""

    _destruct_callbacks = []  # List of destruction callbacks
    _log_list = []  # Store the logs
    _variables = {}
    _infos = IndexedDict()
    _instrument = None  # Product instrument
    _logger = None
    _targets = []

    def __init__(self, manager=None, instrument=None, targets=[],
                 **kwargs):
        if manager is None:
            raise ValueError("A product has to be created with a"
                             " manager.")
        self._manager = manager
        self._instrument = instrument
        self._targets = targets

        for name, value in kwargs.items():
            self.__setattr__(name, value)

    @property
    def index(self):
        return self._manager.get_product_index(self.name)

    @property
    def name(self):
        return self._manager.get_product_name(self)

    @property
    def logger(self):
        if self._logger is None:
            self._logger = self._manager.logger.getChild(self.name)
            self._logger.setLevel(self._manager.logger.getEffectiveLevel())
            log_to_list(self._logger, self._log_list)
        return self._logger

    @property
    def manager(self):
        return self._manager

    @property
    def instrument(self):
        return self._instrument

    @property
    def info(self):
        """Print general informations about the product, in a yaml format."""
        info_dict = self._infos.copy()

        info_dict['variables'] = self._variables

        inst = dict()
        inst['class'] = self._instrument.__class__.__name__
        inst['id'] = self._instrument._identifier
        inst['properties'] = self._instrument.properties
        info_dict['instrument'] = inst

        dest = dict()
        dest['number'] = len(self._destruct_callbacks)
        dest['functions'] = [i.__name__ for i in self._destruct_callbacks]
        info_dict['destruction_callbacks'] = dest

        dest['history'] = self._log_list

        return info_dumper(info_dict)

    @property
    def log(self):
        """Print the log of the product."""
        return "\n".join(self._log_list)

    @property
    def targets(self):
        return copy.deepcopy(self._targets)

    def get_value(self, name):
        """Get a variable value."""
        return self._variables[name]

    def set_value(self, name, value):
        """Set a value to a variable."""
        self._variables[name] = value

    def add_info(self, session, info_dict):
        """Add information dictionaries to the product."""
        if session in self._infos.keys():
            self.logger.warn(f'Session {session} already exists in this '
                             'product infos. Overwriting it.')
        if session in ['variables', 'history', 'destruction_callbacks']:
            self.logger.warn(f'{session} is a protected name of session. '
                             'Skipping.')
            return
        self._infos[session] = info_dict

    def add_target(self, name):
        """Add a stage to the target list."""
        if name not in self._targets:
            self._targets.append(name)
        else:
            self.logger.debug(f'Stage {name} already in product targets.')

    def del_target(self, name):
        """Remove a stage from the target list."""
        if name in self._targets:
            self._targets.remove(name)
        else:
            self.logger.debug(f'Stage {name} not in the target list.')

    def add_destruct_callback(self, callback, *args, **kwargs):
        """Add a destruction callback."""
        func = partial(callback, self, *args, **kwargs)
        self._destruct_callbacks.append(func)

    def destruct(self):
        """Execute the destruction callbacks sequentially."""
        for i, func in enumerate(self._destruct_callbacks):
            try:
                func()
            except Exception as e:
                logger.debug(f"Destruction callback {i} problem. Error: {e}")


class Stage(abc.ABC):
    """Stage process (sub-part) of a pipeline."""

    _default_config = Config()  # stage default config
    _requested_functions = []  # Instrument needed functions
    _required_variables = []  # Product needed variables
    _optional_variables = []  # Optional variables for product ruunning
    _provided = []  # Product variables provided by stage
    _status = 'idle'
    _raise_error = False
    # TODO: Implement instrument compatibility checking

    def __init__(self, factory):
        self._factory = factory

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, stat):
        if stat not in ['idle', 'running', 'done', 'error']:
            raise ValueError(f'Status {stat} not allowed.')
        self._status = stat

    @property
    def name(self):
        return self.factory.get_stage_name(self)

    @property
    def index(self):
        return self.factory.get_stage_index(self.name)

    @property
    def factory(self):
        return self._factory

    @property
    def logger(self):
        return self.factory.logger

    @property
    def defaults(self):
        return copy.deepcopy(Config(**self._default_config))

    def wait(self):
        """Wait process to finish."""
        raise RuntimeError('No processing running in parallel now. '
                           'This should not be possible!')

    def get_variables(self):
        """Return a dict of all needed and optional variables."""
        variables = {}
        for v in self._required_variables:
            variables[v] = self.factory.get_value(self, v)
        # Optional variables will get None if not available
        for v in self._optional_variables:
            try:
                variables[v] = self.factory.get_value(self, v)
            except Exception as e:
                variables[v] = None
                logger.warn('Variable %s not found due to %s.'
                            ' Falling back to None.', v, e)

        return variables

    def get_instrument(self):
        """Get the instrument class of active product."""
        return self.factory.get_instrument()

    @abc.abstractstaticmethod
    def callback(instrument, variables, config, logger):
        """Run the stage. Return a dict of processed variables."""
        # Avoiding access to class stuff make possible parallelizing

    def _call_pipeline(self, instrument, config):
        try:

            # Ensure get all variables before running
            variables = self.get_variables()
            owned = self.factory.owned_variables(self)
            result = {i: None for i in owned}
            conf = self.defaults
            conf.unfreeze()
            conf.update(config)
            conf.freeze()

            self.status = 'running'
            instrument = self.get_instrument()
            self.logger.info(f'Executing {self.name} stage.')
            self.logger.debug(f'Stage {self.name} variables: {variables}')
            self.logger.debug(f'Stage {self.name} config: {conf}')
            result.update(self.callback(instrument, variables, conf,
                                        self.logger))
            self.status = 'done'
            self.logger.debug(f'Stage {self.name} results: {result}')
            for i, v in result.items():
                if i in owned:
                    self.factory.set_value(self, i, v)
        except Exception as e:
            if self._raise_error:
                raise e
            self.logger.error(f'Stage {self.name} error: {e}')
            self.status = 'error'


class Factory():
    _stages = IndexedDict()  # Stages list
    _register = IndexedDict()  # Variables registering
    _active_prod = None
    _active_config = Config()
    _logger = None

    # Targets are defined by product
    """Class to handle execution and product/stage interface."""
    def __init__(self, manager):
        self._manager = manager

    @property
    def logger(self):
        if self._logger is None:
            self._logger = self._manager.logger.getChild('factory')
            self._logger.setLevel(self._manager.logger.getEffectiveLevel())
            self._factory_logger = self._logger
        return self._logger

    def get_value(self, stage, variable):
        """Get a value from a registered variable."""
        if variable not in self._register.keys():
            raise KeyError(f'Variable {variable} not registered.')

        if stage is None or self._register[variable] is None:
            # Non-stage variable access
            return copy.deepcopy(self._active_prod.get_value(variable))

        if stage.status != 'idle':
            raise RuntimeError('Only idle stages can access variables. '
                               f'Current status of {stage.name}'
                               f' stage: {stage.status}')

        # Check the status of registered stage for this variable
        var_stage = self._stages[self._register[variable]]

        if var_stage.status == 'idle':
            # If not started yet, run it
            self.logger.debug(f'Variable {variable} not executed. '
                              f'Opening {var_stage.name} stage to do it.')
            self.run_stage(var_stage.name)

        if var_stage.status == 'running':
            # Wait if already running
            var_stage.wait()

        # Done results
        if var_stage.status == 'done':
            return copy.deepcopy(self._active_prod.get_value(variable))
        elif var_stage.status == 'error':
            raise RuntimeError('Variable {} not accessible due to '
                               '{} stage processing error.')
        else:
            raise ValueError('This should be impossible! Stage status: '
                             f'{var_stage.status}')

    def set_value(self, stage, variable, value):
        """Set a variable value to product."""
        if isinstance(stage, Product):
            self.logger.debug(f'Setting {variable} variable manually for '
                              f'{stage.name} product.')
            stage.set_value(variable, copy.deepcopy(value))
            self._register[variable] = None
            return

        if variable not in self._register.keys():
            raise ValueError(f'Variable {variable} not registered.')

        if isinstance(stage, Stage):
            # If a stage instance is passed, get its name for checking
            stage = stage.name

        if stage != self._register[variable]:
            raise ValueError(f'Stage {stage} not own {variable} variable.')

        var_stage = self._stages[stage]
        if var_stage.status != 'done':
            raise RuntimeError('Only done stages can set variables. '
                               f'Current {var_stage.name} '
                               f'stage status: {var_stage.status}')

        self.logger.debug(f'Setting {variable} variable with value {value} '
                          f'from {var_stage.name} stage.')
        self._active_prod.set_value(variable, copy.deepcopy(value))

    def get_instrument(self):
        """Get the product instrument."""
        # Return a copy, so it cannot be changed
        return copy.deepcopy(self._active_prod.instrument)

    def register_stage(self, name, stage, disable_variables=[]):
        """Register a stage."""
        if stage in self._stages.values():
            raise ValueError('Stage already registered.')

        if name in self._stages.keys():
            raise ValueError(f'Stage name {name} already in use.')

        var = [i for i in stage._provided if i not in disable_variables]
        self.logger.debug(f'Registering stage {stage.name} with {var} '
                          'variables.')
        for i in var:
            self._register[i] = name

        self._stages[name] = stage

    def unregister_stage(self, stage):
        """Remove a stage from the registers."""
        if isinstance(stage, Stage):
            # If a stage instance is passed, get its name for checking
            stage = stage.name

        self.logger.debug(f'Unregistering stage {stage} with'
                          f' {self.owned_variables(stage)} variables.')

        for i, v in self._register.items():
            if v == stage:
                self._register.pop(i)

        self._stages.popitem(stage)

    def get_stage_name(self, instance):
        """Get the name of a stage instance."""
        if instance not in self._stages.values():
            if instance.factory == self:
                return 'unregistered_stage'
            else:
                raise ValueError('Stage not associated to '
                                 'this manager.')

        for i, v in self._stages.items():
            if v == instance:
                return i

    def owned_variables(self, stage):
        """Return the variables owned by a stage."""
        if isinstance(stage, Stage):
            stage = stage.name

        owned = [k for k, v in self._register.items() if v == stage]
        self.logger.debug(f'Stage {stage} owns variables: {owned}')
        return owned

    def activate_product(self, product):
        """Activate a product to this factory."""
        self.reset()
        self.logger.info(f'Actiavting {product.name} product.')
        self._active_prod = product
        self._factory_logger = self._logger
        self._logger = product.logger

    def reset(self):
        """Cleanup all needed informations from this factory."""
        self.logger.info("Reseting factory")
        self._active_prod = None
        self._active_config = Config()
        for v in self._stages.values():
            v.status = 'idle'
        self._logger = self._factory_logger

    def dump_defaults(self):
        """Dump the default configurations in yaml format."""
        conf = {}
        for i, v in self._stages:
            conf[i] = dict(v.config)

        return yaml.dump(conf)

    def run_stage(self, stage):
        """Run a single stage for the active product."""
        if self._active_prod is None:
            raise ValueError("No product has been activated!")

        self.logger.info(f'Opening job for {self._stages[stage].name}'
                         f' stage for {self._active_prod.name} product.')

        stage_conf = self._active_config.get(stage, {})
        stage_conf = Config(stage_conf)
        instrument = self._active_prod.instrument

        self.logger.debug('Freezing instrument and config.')
        stage_conf.freeze()
        instrument.freeze()

        # Execute!
        self._stages[stage]._call_pipeline(instrument, stage_conf)

        if self._stages[stage].status == 'error':
            self.logger.error(f'Stage {stage} failed.')
        elif self._stages[stage].status == 'done':
            self.logger.info(f'Stage {stage} done without problems.')

        self.logger.debug('Unfreezing instrument and config')
        stage_conf.unfreeze()
        instrument.unfreeze()
        self.logger.info(f'Closing job for {stage} stage.')

    def run(self, config, target=None):
        """Run te factory to the product."""
        self._active_config = Config(config)
        targets = target or self._active_prod.targets
        if not check_iterable(targets):
            targets = [targets]
        self.logger.info(f'Executing pipeline with {targets} targets.')
        for i in targets:
            self.run_stage(i)
        self.logger.info('Pipeline executed.')
        self.reset()


class Manager(abc.ABC):
    _config = Config()
    _products = IndexedDict()
    _factory = None
    _logger = None

    # TODO: handle configs
    """Class to handle the general pipeline management."""
    def __init__(self):
        self._factory = Factory(self)
        self._config['stages'] = Config()

    @property
    def factory(self):
        return self._factory

    @property
    def logger(self):
        if self._logger is None:
            self._logger = logger.getChild('manager')
            self._logger.setLevel(logger.getEffectiveLevel())
        return self._logger

    @property
    def config(self):
        return copy.deepcopy(self._config)

    @abc.abstractmethod
    def setup_pipeline(self, *args, **kwargs):
        """Set up the pipeline, register stages, etc.

        Can be used to read a config.
        """

    @abc.abstractmethod
    def setup_products(self, *args, **kwargs):
        """Set up products based on specified parameters.

        Examples:
            - Read files from a folder, with filters
            - List files
            - Load a configure file
        """

    def add_product(self, name, product, index=None, requires=[]):
        """Add a new product to this Manager.

        Parameters:
            name : string
                Indentifying name for this product. Uniqueness checked.
            product : `Product`
                Product valid instance.
            index : int
                Insert the product in a given index.
            requires : list(string)
                List of products to be processed first.
        """
        index1 = None

        inds = [self._products.index(r)
                for r in requires]
        if len(inds) > 0:
            index1 = max(inds)

        if index1 is not None and index is not None:
            if index1 < index:
                logger.warn(f'Cannot insert product {index} before its'
                            ' requirements.'
                            f' Overwriting to {index1}')
                index = index1

        if index is None:
            # If indexes are not set, just append
            self._products[name] = product
        else:
            self._products.insert_at(index, name, product)

    def get_product_index(self, name):
        """Return the current index of a given product."""
        for i, k in enumerate(self._products.keys()):
            if k == name:
                return i

        self.logger.warn(f"Product {name} not found.")

    def get_product_name(self, instance):
        """Return the name of a product based on its instance."""
        if instance not in self._products.values():
            if instance.manager == self:
                return 'unregistered_product'
            else:
                raise ValueError('Product not associated to '
                                 'this manager.')
        for i, v in self._products.items():
            if v == instance:
                return i

        self.logger.warn(f"No product associated to {str(instance)} instance.")

    def del_product(self, name):
        """Remove and clean a product."""
        try:
            del self._products[name]
        except KeyError:
            logger.debug(f"Product {name} not in this factory.")
            pass

    def register_stage(self, name, stage, disable_variables=[]):
        """Register a stage."""
        self.factory.register_stage(name, stage, disable_variables)

    def unregister_stage(self, name):
        """Remove a stage from the registers."""
        self.factory.unregister_stage(name)

    def set_value(self, product, variable, value):
        """Set a default value to a variable not owned by a stage."""
        self.factory.set_value(product, variable, value)

    def get_value(self, product, variable):
        """Get a default value from a variable not owned by a stage."""
        return self.factory.get_value()

    def show_products(self):
        """Show the created products in a list."""
        if len(self._products) == 0:
            print('# No products on this manager.')
        else:
            for i, n in enumerate(self._products.keys()):
                print(f"{i}\t{n}")

    def run(self, index=None, target=None):
        if index is not None:
            if not check_iterable(index):
                index = [index]
        else:
            index = list(range(len(self._products)))

        n = len(index)
        self.logger.info(f'Processing {n} products.')

        for i in index:
            self.logger.info(f"Processing product {i+1} from {n}")
            name = list(self._products.keys())[i]
            self.factory.activate_product(self._products[name])
            self.factory.run(self.config['stages'], target=target)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Base classes for astronomical catalogs query."""

import copy
import abc
import numpy as np
from astropy.table import Table
from astropy.coordinates import Angle, SkyCoord, match_coordinates_sky

from ..logger import logger
from ._online_tools import _timeout_retry, get_center_radius, \
                           _wrap_query_table
from ..astrometry.coords_utils import guess_coordinates


__all__ = ['match_indexes']


def match_indexes(ra, dec, cat_ra, cat_dec, limit_angle):
    """Match list of coordinates with a calatog.

    Parameters
    ----------
    ra, dec: list of float
        List of objects coordinates to be matched in the catalog. All
        coordinates must be in decimal degrees.
    cat_ra, cat_dec: list of float
        List of catalog coordinates. All coordinates must be in decimal
        degrees.
    limit_angle: string, float, `~astropy.coordinates.Angle`
        Angle limit for matching indexes. If string, if must be
        `~astropy.coordinates.Angle` compatible. If float, it will be
        interpreted as a decimal degree.
    logger: `~logging.Logger`

    Returns
    -------
    index: `~numpy.ndarray`
        List containing the indexes in the catalog that matched the object
        coordinates. If -1, it represents objects not matched.
    """
    # # Matching using astropy's skycoord can be slow for big catalogs
    ind, dist, _ = match_coordinates_sky(SkyCoord(ra, dec, unit=('degree',
                                                                 'degree'),
                                                  frame='icrs'),
                                         SkyCoord(cat_ra, cat_dec,
                                                  unit=('degree', 'degree'),
                                                  frame='icrs'))

    index = np.zeros(len(ra), dtype=np.int)
    index.fill(-1)   # a nan index

    lim = Angle(limit_angle)
    for k in range(len(ra)):
        if dist[k] <= lim:
            index[k] = ind[k]

    logger.debug('Matched %s objects in catalog from %s total',
                 np.sum(index != -1), len(ra))

    return index


class _BaseCatalog(abc.ABC):
    """Base class for catalog query."""

    # stores all the needed information to avoid continuous redundat queries
    _last_query_info = None
    _last_query_table = None
    comment = None

    def __evaluate__(self, center, radius=None, **kwargs):
        """Query the results in the catalog.

        Parameters:
        -----------
        center: string or tuple
            The center of the search field.
            If center is a string, can be an object name or the string
            containing the object coordinates. If it is a tuple, have to be
            (ra, dec) coordinates, in hexa or decimal degrees format.
        radius: string, float, `~astropy.coordinates.Angle`
                or None (optional)
            The radius to search. If None, the query will be performed as
            single object query mode. Else, the query will be performed as
            field mode. If a string value is passed, it must be readable by
            astropy.coordinates.Angle. If a float value is passed, it will
            be interpreted as a decimal degree radius.
        ** kwargs
            other arguments to be passed to the catalog functions.
        """
        if radius is None:
            return self.query_object(center, **kwargs)
        return self.query_region(center, radius, **kwargs)

    def flush(self):
        """Clear previous query informations."""
        del self._last_query_info
        del self._last_query_table

    def copy(self):
        """Get a new copy of the catalog."""
        return copy.deepcopy(self)


class _BasePhotometryCatalog(_BaseCatalog, abc.ABC):
    """A base class for photometry catalogs."""

    name = None
    type = None
    flux_unit = None
    available_filters = []
    bibcode = None

    def check_filter(self, band, raise_error=True):
        """Check if a filter is available for this catalog."""
        if not raise_error:
            return band in self.available_filters
        if band in self.available_filters:
            return True

        raise ValueError(f'This catalog does not support {band} filter. '
                         'The available formats are:'
                         f' {self.available_filters}')

    def filter_ra_dec(self, query):
        """Filter coordinates in a query result."""
        if self.ra_key is None or self.dec_key is None:
            raise ValueError("Invalid RA or Dec keys.")

        if query is None:
            query = self._last_query_table

        ra = query[self.ra_key].data
        dec = query[self.dec_key].data

        # Solve the most common types of coordinates
        coords = guess_coordinates(ra, dec)
        ra = np.array(coords.ra.degree)
        dec = np.array(coords.dec.degree)

        return ra, dec

    def filter_id(self, query=None):
        """Filter object names in a query result."""
        if self.id_key is None:
            raise ValueError("Invalid ID key.")

        if query is None:
            query = self._last_query_table

        if self.id_key == -1:
            return np.array(['']*len(query))

        idn = query[self.id_key].data
        return self._id_resolve(idn)

    def _query(self, querier, *args, **kwargs):
        """Perform a query. All args are passed to querier."""
        query_info = (args, kwargs)
        if (query_info == self._last_query_info and
           self._last_query_table is not None):
            logger.debug("Loading cached query.")
            return copy.copy(self._last_query_table)

        logger.info("Performing %s query with parameters: %s",
                    self.__class__.__name__, query_info)

        self._last_query_info = query_info
        self._last_query_table = None

        query = _timeout_retry(querier, *args, **kwargs)
        if query is None:
            raise RuntimeError("No online catalog result found.")
        self._last_query_table = _wrap_query_table(query)
        return copy.copy(self._last_query_table)

    def _match_objects(self, ra, dec, band, limit_angle,
                       flux_keys, table_props):
        c_ra, c_dec, radius = get_center_radius(ra, dec)
        query = self.query_region((c_ra, c_dec), radius)
        cat = Table()
        cat['id'] = self.filter_id(query)
        cat['ra'], cat['dec'] = self.filter_ra_dec(query)

        indexes = match_indexes(ra, dec, cat['ra'], cat['dec'], limit_angle)
        if band is not None:
            for k, v in zip(flux_keys, self.filter_flux(band, query)):
                cat[k] = v
        else:
            for k in flux_keys:
                cat[k] = np.full(len(cat), fill_value=np.nan)

        res = Table()
        res['ra'] = ra
        res['dec'] = dec
        for k, empty in table_props:
            res['cat_'+k] = [cat[k][i] if i < 0 else empty for i in indexes]

        return res



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import numpy as np
import pytest

from astropop.image.imarith import imarith
from astropop.framedata import FrameData
from astropop.math.physical import UnitsError, units, QFloat
from astropop.testing import *


# TODO: % and ** functions


SHAPE = (10, 10)


def gen_frame(value, uncertainty=None, unit='adu'):
    """Gen frames with {'v', 'u'} dict"""
    shape = SHAPE
    frame = FrameData(np.ones(shape, dtype='f8'), unit=unit,
                      uncertainty=uncertainty)
    frame.data[:] = value
    return frame


def std_test_frame(op, frame1, frame2, result, inplace, handle_mask):
    """Standart comparison tests with frames."""
    if handle_mask:
        # Use two different masks to check if they are compined
        mask1 = np.zeros(SHAPE, dtype=bool)
        mask1[2, 2] = 1
        frame1.mask = mask1

        exp_mask = np.zeros(SHAPE, dtype=bool)
        exp_mask[2, 2] = 1

        if isinstance(frame2, FrameData):
            # If frame2 is qfloat, quantity or a number, it don't have mask
            mask2 = np.zeros(SHAPE, dtype=bool)
            mask2[3, 3] = 1
            exp_mask[3, 3] = 1
            frame2.mask = mask2

        result.mask = exp_mask

    res = imarith(frame1, frame2, op, inplace=inplace,
                  join_masks=handle_mask)

    assert_equal(res.data, result.data)
    assert_almost_equal(res.get_uncertainty(False),
                        result.get_uncertainty(False))
    assert_equal(res.unit, result.unit)
    if handle_mask:
        assert_equal(res.mask, result.mask)

    if inplace:
        assert_is(res, frame1)
    else:
        assert_is_not(res, frame1)


@pytest.mark.parametrize('handle_mask', [True, False])
@pytest.mark.parametrize('inplace', [True, False])
class Test_Imartih_OPs_add:
    op = '+'

    def test_uncertainty_all_none(self, inplace, handle_mask):
        frame1 = gen_frame(30, None, unit='adu')
        frame2 = gen_frame(10, None, unit='adu')
        result = gen_frame(40, None, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_none(self, inplace, handle_mask):
        frame1 = gen_frame(30, None, unit='adu')
        frame2 = gen_frame(10, 0, unit='adu')
        result = gen_frame(40, 0, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_none(self, inplace, handle_mask):
        frame1 = gen_frame(30, 0, unit='adu')
        frame2 = gen_frame(10, None, unit='adu')
        result = gen_frame(40, 0, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_set(self, inplace, handle_mask):
        frame1 = gen_frame(30, 3, unit='adu')
        frame2 = gen_frame(10, None, unit='adu')
        result = gen_frame(40, 3, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_set(self, inplace, handle_mask):
        frame1 = gen_frame(30, 0, unit='adu')
        frame2 = gen_frame(10, 3, unit='adu')
        result = gen_frame(40, 3, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_all_set(self, inplace, handle_mask):
        frame1 = gen_frame(30, 3, unit='adu')
        frame2 = gen_frame(10, 4, unit='adu')
        result = gen_frame(40, 5, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_zero(self, inplace, handle_mask):
        frame1 = gen_frame(10, unit='adu')
        result = gen_frame(10, unit='adu')

        # pure number should fail
        with pytest.raises(UnitsError):
            std_test_frame(self.op, frame1, 0, result,
                           inplace=inplace, handle_mask=handle_mask)

        # quantity must pass
        std_test_frame(self.op, frame1, 0*units.adu, result,
                       inplace=inplace, handle_mask=handle_mask)

        # qfloat must pass
        std_test_frame(self.op, frame1, QFloat(0, unit='adu'), result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_one(self, inplace, handle_mask):
        frame1 = gen_frame(10, unit='adu')
        result = gen_frame(11, unit='adu')

        # pure number should fail
        with pytest.raises(UnitsError):
            std_test_frame(self.op, frame1, 1, result,
                           inplace=inplace, handle_mask=handle_mask)

        # quantity must pass
        std_test_frame(self.op, frame1, 1*units.adu, result,
                       inplace=inplace, handle_mask=handle_mask)

        frame1 = gen_frame(10, unit='adu')  # if not do this, result gets wrong on inplace
        # qfloat must pass
        std_test_frame(self.op, frame1, QFloat(1, unit='adu'), result,
                       inplace=inplace, handle_mask=handle_mask)


@pytest.mark.parametrize('handle_mask', [True, False])
@pytest.mark.parametrize('inplace', [True, False])
class Test_Imartih_OPs_sub:
    op = '-'

    def test_uncertainty_all_none(self, inplace, handle_mask):
        frame1 = gen_frame(30, None, unit='adu')
        frame2 = gen_frame(10, None, unit='adu')
        result = gen_frame(20, None, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_none(self, inplace, handle_mask):
        frame1 = gen_frame(30, None, unit='adu')
        frame2 = gen_frame(10, 0, unit='adu')
        result = gen_frame(20, 0, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_none(self, inplace, handle_mask):
        frame1 = gen_frame(30, 0, unit='adu')
        frame2 = gen_frame(10, None, unit='adu')
        result = gen_frame(20, 0, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_set(self, inplace, handle_mask):
        frame1 = gen_frame(30, 3, unit='adu')
        frame2 = gen_frame(10, None, unit='adu')
        result = gen_frame(20, 3, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_set(self, inplace, handle_mask):
        frame1 = gen_frame(30, 0, unit='adu')
        frame2 = gen_frame(10, 3, unit='adu')
        result = gen_frame(20, 3, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_all_set(self, inplace, handle_mask):
        frame1 = gen_frame(30, 3, unit='adu')
        frame2 = gen_frame(10, 4, unit='adu')
        result = gen_frame(20, 5, unit='adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_zero(self, inplace, handle_mask):
        frame1 = gen_frame(10, unit='adu')
        result = gen_frame(10, unit='adu')

        # pure number should fail
        with pytest.raises(UnitsError):
            std_test_frame(self.op, frame1, 0, result,
                           inplace=inplace, handle_mask=handle_mask)

        # quantity must pass
        std_test_frame(self.op, frame1, 0*units.adu, result,
                       inplace=inplace, handle_mask=handle_mask)

        # qfloat must pass
        std_test_frame(self.op, frame1, QFloat(0, unit='adu'), result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_one(self, inplace, handle_mask):
        frame1 = gen_frame(10, unit='adu')
        result = gen_frame(9, unit='adu')

        # pure number should fail
        with pytest.raises(UnitsError):
            std_test_frame(self.op, frame1, 1, result,
                           inplace=inplace, handle_mask=handle_mask)

        # quantity must pass
        std_test_frame(self.op, frame1, 1*units.adu, result,
                       inplace=inplace, handle_mask=handle_mask)

        frame1 = gen_frame(10, unit='adu')  # if not do this, result gets wrong on inplace
        # qfloat must pass
        std_test_frame(self.op, frame1, QFloat(1, unit='adu'), result,
                       inplace=inplace, handle_mask=handle_mask)


@pytest.mark.parametrize('handle_mask', [True, False])
@pytest.mark.parametrize('inplace', [True, False])
class Test_Imartih_OPs_mul:
    op = '*'

    def test_uncertainty_all_none(self, inplace, handle_mask):
        frame1 = gen_frame(5, None, unit='adu')
        frame2 = gen_frame(6, None, unit='electron/adu')
        result = gen_frame(30, None, unit='electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_none(self, inplace, handle_mask):
        frame1 = gen_frame(5, None, unit='adu')
        frame2 = gen_frame(6, 0, unit='electron/adu')
        result = gen_frame(30, 0, unit='electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_none(self, inplace, handle_mask):
        frame1 = gen_frame(5, 0, unit='adu')
        frame2 = gen_frame(6, None, unit='electron/adu')
        result = gen_frame(30, 0, unit='electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_set(self, inplace, handle_mask):
        frame1 = gen_frame(5, 0.3, unit='adu')
        frame2 = gen_frame(6, None, unit='electron/adu')
        result = gen_frame(30, 1.8, unit='electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_set(self, inplace, handle_mask):
        frame1 = gen_frame(5, 0, unit='adu')
        frame2 = gen_frame(6, 0.3, unit='electron/adu')
        result = gen_frame(30, 1.5, unit='electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_all_set(self, inplace, handle_mask):
        frame1 = gen_frame(5, 0.5, unit='adu')
        frame2 = gen_frame(6, 0.3, unit='electron/adu')
        result = gen_frame(30, 3.35410196624, unit='electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_zero(self, inplace, handle_mask):
        frame1 = gen_frame(10, unit='adu')
        result = gen_frame(0, unit='adu')

        # pure number must pass
        std_test_frame(self.op, frame1, 0, result,
                       inplace=inplace, handle_mask=handle_mask)

        # quantity must pass
        std_test_frame(self.op, frame1, 0*units.dimensionless_unscaled, result,
                       inplace=inplace, handle_mask=handle_mask)

        # qfloat must pass
        std_test_frame(self.op, frame1, QFloat(0), result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_one(self, inplace, handle_mask):
        frame1 = gen_frame(10, unit='adu')
        result = gen_frame(10, unit='adu')

        # pure number must pass
        std_test_frame(self.op, frame1, 1, result,
                       inplace=inplace, handle_mask=handle_mask)

        # quantity must pass
        std_test_frame(self.op, frame1, 1*units.dimensionless_unscaled,
                       result,
                       inplace=inplace, handle_mask=handle_mask)

        frame1 = gen_frame(10)  # if not do this, result gets wrong on inplace
        # qfloat must pass
        std_test_frame(self.op, frame1, QFloat(1), result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_unit_change(self, inplace, handle_mask):
        # FrameData must pass
        frame1 = gen_frame(10, unit='adu')
        frame2 = gen_frame(2, unit='electron/adu')
        result = gen_frame(20, unit='electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

        # Quantity must pass
        frame1 = gen_frame(10, unit='adu')
        frame2 = 2*units.electron/units.adu
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

        # QFloat must pass
        frame1 = gen_frame(10, unit='adu')
        frame2 = QFloat(2, unit='electron/adu')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)


@pytest.mark.parametrize('handle_mask', [True, False])
@pytest.mark.parametrize('inplace', [True, False])
class Test_Imartih_OPs_div:
    op = '/'

    def test_uncertainty_all_none(self, inplace, handle_mask):
        frame1 = gen_frame(10, None, 'adu')
        frame2 = gen_frame(2, None, 'electron')
        result = gen_frame(5, None, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_none(self, inplace, handle_mask):
        frame1 = gen_frame(10, None, unit='adu')
        frame2 = gen_frame(2, 0, unit='electron')
        result = gen_frame(5, 0, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_none(self, inplace, handle_mask):
        frame1 = gen_frame(10, 0, unit='adu')
        frame2 = gen_frame(2, None, unit='electron')
        result = gen_frame(5, 0, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_set(self, inplace, handle_mask):
        frame1 = gen_frame(10, 0.2, unit='adu')
        frame2 = gen_frame(2, None, unit='electron')
        result = gen_frame(5, 0.1, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_set(self, inplace, handle_mask):
        frame1 = gen_frame(10, None, unit='adu')
        frame2 = gen_frame(2, 0.2, unit='electron')
        result = gen_frame(5, 0.5, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_all_set(self, inplace, handle_mask):
        frame1 = gen_frame(10, 0.1, unit='adu')
        frame2 = gen_frame(2, 0.2, unit='electron')
        result = gen_frame(5, 0.502493781056, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_one(self, inplace, handle_mask):
        frame1 = gen_frame(10, unit='adu')
        result = gen_frame(10, unit='adu')

        # pure number must pass
        std_test_frame(self.op, frame1, 1, result,
                       inplace=inplace, handle_mask=handle_mask)

        # quantity must pass
        std_test_frame(self.op, frame1, 1*units.dimensionless_unscaled,
                       result,
                       inplace=inplace, handle_mask=handle_mask)

        frame1 = gen_frame(10)  # if not do this, result gets wrong on inplace
        # qfloat must pass
        std_test_frame(self.op, frame1, QFloat(1), result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_unit_change(self, inplace, handle_mask):
        frame1 = gen_frame(10, unit='adu')
        result = gen_frame(5, unit='electron')

        # FrameData must pass
        frame2 = gen_frame(2, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

        # Quantity must pass
        frame1 = gen_frame(10)
        frame2 = 2*units.adu/units.electron
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

        # QFloat must pass
        frame1 = gen_frame(10)
        frame2 = QFloat(2, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)


@pytest.mark.parametrize('handle_mask', [True, False])
@pytest.mark.parametrize('inplace', [True, False])
class Test_Imartih_OPs_fllordiv:
    op = '//'

    def test_uncertainty_all_none(self, inplace, handle_mask):
        frame1 = gen_frame(11, None, unit='adu')
        frame2 = gen_frame(2, None, unit='s')
        result = gen_frame(5, None, unit='adu/s')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_none(self, inplace, handle_mask):
        frame1 = gen_frame(11, None, unit='adu')
        frame2 = gen_frame(2, 0, unit='s')
        result = gen_frame(5, 0, unit='adu/s')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_none(self, inplace, handle_mask):
        frame1 = gen_frame(11, 0, unit='adu')
        frame2 = gen_frame(2, None, unit='s')
        result = gen_frame(5, 0, unit='adu/s')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f1_set(self, inplace, handle_mask):
        frame1 = gen_frame(11, 0.2, unit='adu')
        frame2 = gen_frame(2, None, unit='s')
        result = gen_frame(5, 0.0, unit='adu/s')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_f2_set(self, inplace, handle_mask):
        frame1 = gen_frame(11, None, unit='adu')
        frame2 = gen_frame(2, 0.2, unit='s')
        result = gen_frame(5, 0.0, unit='adu/s')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_uncertainty_all_set(self, inplace, handle_mask):
        frame1 = gen_frame(11, 0.1, unit='adu')
        frame2 = gen_frame(2, 0.2, unit='s')
        result = gen_frame(5, 0, unit='adu/s')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_one(self, inplace, handle_mask):
        frame1 = gen_frame(10.5, unit='adu')
        result = gen_frame(10, unit='adu')

        # pure number must pass
        std_test_frame(self.op, frame1, 1, result,
                       inplace=inplace, handle_mask=handle_mask)

        # quantity must pass
        std_test_frame(self.op, frame1, 1*units.dimensionless_unscaled,
                       result,
                       inplace=inplace, handle_mask=handle_mask)

        frame1 = gen_frame(10)  # if not do this, result gets wrong on inplace
        # qfloat must pass
        std_test_frame(self.op, frame1, QFloat(1), result,
                       inplace=inplace, handle_mask=handle_mask)

    def test_unit_change(self, inplace, handle_mask):
        frame1 = gen_frame(11, unit='adu')
        result = gen_frame(5, unit='electron')

        # FrameData must pass
        frame2 = gen_frame(2, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

        # Quantity must pass
        frame1 = gen_frame(10)
        frame2 = 2*units.adu/units.electron
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)

        # QFloat must pass
        frame1 = gen_frame(10, unit='adu')
        frame2 = QFloat(2, unit='adu/electron')
        std_test_frame(self.op, frame1, frame2, result,
                       inplace=inplace, handle_mask=handle_mask)


def test_invalid_op():
    frame1 = FrameData(np.zeros((10, 10)), unit='')
    frame2 = FrameData(np.zeros((10, 10)), unit='')
    with pytest.raises(ValueError) as exc:
        imarith(frame1, frame2, 'not an op')
        assert_in('not supported', str(exc.value))


def test_invalid_shapes():
    frame1 = FrameData(np.zeros((10, 10)), unit='')
    frame2 = FrameData(np.zeros((5, 5)), unit='')
    with pytest.raises(ValueError):
        imarith(frame1, frame2, '+')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
from astropop.logger import logger, log_to_list, resolve_level_string, \
                            ListHandler
from astropop.testing import *


class Test_Logger_To_List():
    @pytest.mark.parametrize('level, expected', [('WARN', 2), ('INFO', 3),
                                                ('DEBUG', 4)])
    def test_logger_list_defaults(self, level, expected):
        mylog = logger.getChild('testing')
        logs = []
        log_to_list(mylog, logs)
        mylog.setLevel(level)
        mylog.error('Error test')
        mylog.warning('Warning test')
        mylog.info('Info test')
        mylog.debug('Debug test')
        assert_equal(mylog.name, 'astropop.testing')
        assert_equal(len(logs), expected)
        for i, k in zip(['Error test', 'Warning test', 'Info test',
                        'Debug test'][0:expected],
                        logs):
            assert_equal(i, k)

    @pytest.mark.parametrize('level, expected', [('WARN', 2), ('INFO', 3),
                                                ('DEBUG', 4)])
    def test_logger_list_only_messagens(self, level, expected):
        mylog = logger.getChild('testing')
        logs = []
        log_to_list(mylog, logs, full_record=False)
        mylog.setLevel(level)
        mylog.error('Error test %i', 40)
        mylog.warning('Warning test %i', 30)
        mylog.info('Info test %i', 20)
        mylog.debug('Debug test %i', 10)
        assert_equal(mylog.name, 'astropop.testing')
        assert_equal(len(logs), expected)
        for i, k in zip(['Error test 40', 'Warning test 30', 'Info test 20',
                         'Debug test 10'][0:expected], logs):
            assert_equal(i, k)

    @pytest.mark.parametrize('level, expected', [('WARN', 2), ('INFO', 3),
                                                ('DEBUG', 4)])
    def test_logger_list_full_record(self, level, expected):
        mylog = logger.getChild('testing')
        logs = []
        log_to_list(mylog, logs, full_record=True)
        mylog.setLevel(level)
        mylog.error('Error test')
        mylog.warning('Warning test')
        mylog.info('Info test')
        mylog.debug('Debug test')
        assert_equal(mylog.name, 'astropop.testing')
        assert_equal(len(logs), expected)
        for i, k, n in zip(['Error', 'Warning', 'Info',
                            'Debug'][0:expected],
                        logs,
                        [40, 30, 20, 10][0:expected]):
            assert_equal(f'{i} test', k.msg)
            assert_equal(k.name, 'astropop.testing')
            assert_equal(k.levelno, n)
            assert_equal(k.levelname, i.upper())


def test_logger_remove_handler():
    mylog = logger.getChild('testing')
    msg = 'Some error happend here.'
    logs = []
    lh = log_to_list(mylog, logs)
    mylog.setLevel('DEBUG')
    mylog.error(msg)
    assert_is_instance(lh, ListHandler)
    assert_in(lh, mylog.handlers)
    mylog.removeHandler(lh)
    assert_not_in(lh, mylog.handlers)
    assert_equal(logs[0], msg)
    assert_equal(lh.log_list[0], msg)
    assert_equal(lh.log_list, logs)


def test_logger_no_loglist():
    mylog = logger.getChild('testing')
    msg = 'Some error happend here.'
    lh = ListHandler()
    assert_is_instance(lh.log_list, list)
    mylog.addHandler(lh)
    mylog.error(msg)
    assert_equal(lh.log_list[0], msg)


def test_logger_list_debug():
    mylog = logger.getChild('testing')
    logs = []
    log_to_list(mylog, logs)
    mylog.setLevel('DEBUG')
    mylog.warning('Warning test')
    mylog.error('Error test')
    mylog.info('Info test')
    mylog.debug('Debug test')
    assert_equal(mylog.name, 'astropop.testing')
    assert_equal(len(logs), 4)


@pytest.mark.parametrize('val, res', [('DEBUG', 10),
                                      ('INFO', 20),
                                      ('WARN', 30),
                                      ('WARNING', 30),
                                      ('ERROR', 40),
                                      ('CRITICAL', 50),
                                      ('FATAL', 50),
                                      (50, 50)])
def test_resolve_string(val, res):
    assert_equal(resolve_level_string(val), res)


def test_invalid_levels_invalid_string():
    with pytest.raises(AttributeError):
        resolve_level_string('NOT_A_LOGLEVEL')


def test_invalid_levels_none():
    with pytest.raises(TypeError):
        resolve_level_string(None)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Data store containers with in-disk caching support and another features."""

from .framedata import FrameData
from .util import read_framedata, check_framedata
from .memmap import MemMapArray
from .compat import imhdus, EmptyDataError


__all__ = ['FrameData', 'read_framedata', 'check_framedata',
           'MemMapArray', 'imhdus', 'EmptyDataError']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
from astropop.math.hasher import hasher
from astropop.math.array import xy2r, iraf_indices, trim_array,  \
                                all_equal
from astropop.math import gaussian, moffat
import numpy as np

from astropop.testing import *


def test_hasher():
    s = 'asdf1234 &*()[]'
    h = hasher(s, 10)
    assert_equal(h, '4b37febb5e')


def test_xy2r():
    f = np.arange(4).reshape((2, 2))
    x, y = iraf_indices(f)
    r, outf = xy2r(x, y, f, 0.0, 0.0)
    assert_equal(r, [0, 1, 1, np.sqrt(2)])
    assert_equal(f.ravel(), outf)  # no reordering expected


def test_iraf_indices():
    f = np.arange(4).reshape((2, 2))
    x, y = iraf_indices(f)
    assert_equal(x, [[0, 1], [0, 1]])
    assert_equal(y, [[0, 0], [1, 1]])

class Test_AllEqual():
    def test_all_equal_1d_true(self):
        a = np.ones(10)
        assert_true(all_equal(a))

        a = np.zeros(10)
        assert_true(all_equal(a))

        a = np.array(['aAa']*10)
        assert_true(all_equal(a))

    def test_all_equal_1d_false(self):
        a = np.ones(10)
        a[2] = 2
        assert_false(all_equal(a))

        a = np.zeros(10)
        a[2] = 2
        assert_false(all_equal(a))

        a = np.array(['aAa']*10)
        a[2] = 'bBb'
        assert_false(all_equal(a))

    def test_all_equal_2d_true(self):
        a = np.ones((5, 5))
        assert_true(all_equal(a))

        a = np.zeros((5, 5))
        assert_true(all_equal(a))

        a = ['aAa']*5
        a = np.array([a]*5)
        assert_true(all_equal(a))

    def test_all_equal_2d_false(self):
        a = np.ones((5, 5))
        a[2][2] = 2
        assert_false(all_equal(a))

        a = np.zeros((5, 5))
        a[2][2] = 2
        assert_false(all_equal(a))

        a = ['aAa']*5
        a = np.array([a]*5)
        a[2][2] = 'bBb'
        assert_false(all_equal(a))


class Test_TrimArray():
    def test_trim_array_centered(self):
        y, x = np.indices((100, 100))
        a = x*y
        ta, tx, ty = trim_array(a, 21, (50, 50), (y, x))
        assert_equal(ta, a[39:61, 39:61])
        assert_equal(tx, x[39:61, 39:61])
        assert_equal(ty, y[39:61, 39:61])
        assert_equal(np.min(tx), 39)
        assert_equal(np.max(tx), 60)
        assert_equal(np.min(ty), 39)
        assert_equal(np.max(ty), 60)
        assert_equal(np.min(ta), 39*39)
        assert_equal(np.max(ta), 60*60)


    def test_trim_array_right(self):
        y, x = np.indices((100, 100))
        a = x*y
        ta, tx, ty = trim_array(a, 21, (95, 50), (y, x))
        assert_equal(ta, a[39:61, 84:])
        assert_equal(tx, x[39:61, 84:])
        assert_equal(ty, y[39:61, 84:])
        assert_equal(np.min(tx), 84)
        assert_equal(np.max(tx), 99)
        assert_equal(np.min(ty), 39)
        assert_equal(np.max(ty), 60)
        assert_equal(np.min(ta), 84*39)
        assert_equal(np.max(ta), 99*60)


    def test_trim_array_left(self):
        y, x = np.indices((100, 100))
        a = x*y
        ta, tx, ty = trim_array(a, 21, (5, 50), (y, x))
        assert_equal(ta, a[39:61, :16])
        assert_equal(tx, x[39:61, :16])
        assert_equal(ty, y[39:61, :16])
        assert_equal(np.min(tx), 0)
        assert_equal(np.max(tx), 15)
        assert_equal(np.min(ty), 39)
        assert_equal(np.max(ty), 60)
        assert_equal(np.min(ta), 0*39)
        assert_equal(np.max(ta), 15*60)


    def test_trim_array_bottom(self):
        y, x = np.indices((100, 100))
        a = x*y
        ta, tx, ty = trim_array(a, 21, (50, 5), (y, x))
        assert_equal(ta, a[:16, 39:61])
        assert_equal(tx, x[:16, 39:61])
        assert_equal(ty, y[:16, 39:61])
        assert_equal(np.min(ty), 0)
        assert_equal(np.max(ty), 15)
        assert_equal(np.min(tx), 39)
        assert_equal(np.max(tx), 60)
        assert_equal(np.min(ta), 0*39)
        assert_equal(np.max(ta), 15*60)


    def test_trim_array_top(self):
        y, x = np.indices((100, 100))
        a = x*y
        ta, tx, ty = trim_array(a, 21, (50, 95), (y, x))
        assert_equal(ta, a[84:, 39:61])
        assert_equal(tx, x[84:, 39:61])
        assert_equal(ty, y[84:, 39:61])
        assert_equal(np.min(ty), 84)
        assert_equal(np.max(ty), 99)
        assert_equal(np.min(tx), 39)
        assert_equal(np.max(tx), 60)
        assert_equal(np.min(ta), 84*39)
        assert_equal(np.max(ta), 99*60)


    def test_trim_array_no_indices(self):
        y, x = np.indices((100, 100))
        a = x*y
        ta, tx, ty = trim_array(a, 21, (50, 95))
        assert_equal(ta, a[84:, 39:61])
        assert_equal(tx, 11)
        assert_equal(ty, 11)


    def test_trim_array_one_origin(self):
        y, x = np.indices((100, 100))
        a = x*y
        ta, tx, ty = trim_array(a, 21, (50, 95), origin=1)
        assert_equal(ta, a[85:, 40:62])
        assert_equal(tx, 11)
        assert_equal(ty, 11)


    def test_trim_array_even(self):
        y, x = np.indices((100, 100))
        a = x*y
        ta, tx, ty = trim_array(a, 20, (50, 95))
        assert_equal(ta, a[85:, 40:61])
        assert_equal(tx, 10)
        assert_equal(ty, 10)


@pytest.mark.skip
class Test_Gaussian:
    def test_gaussian_normalize(self):
        pass

    def test_gaussian_fwhm(self):
        pass

    def test_gaussian_r(self):
        pass

    def test_gaussian_1d(self):
        pass

    def test_gaussian_2d(self):
        pass

    def test_gaussian_1d_model(self):
        pass

    def test_gaussian_2d_model(self):
        pass

    def test_gaussian_r_model(self):
        pass


@pytest.mark.skip
class Test_Moffat:
    def test_moffat_normalize(self):
        pass

    def test_moffat_fwhm(self):
        pass

    def test_moffat_r(self):
        pass

    def test_moffat_1d(self):
        pass

    def test_moffat_2d(self):
        pass

    def test_moffat_1d_model(self):
        pass

    def test_moffat_2d_model(self):
        pass

    def test_moffat_r_model(self):
        pass



"""Module to manage and classify fits files."""

import os
import fnmatch
import glob
from pathlib import Path
import numpy as np

from astropy.io import fits
from astropy.table import Column

from ._db import SQLDatabase, _ID_KEY, sql, SQLTable, SQLColumnMap
from .fits_utils import _fits_extensions, \
                        _fits_extensions_with_compress
from .framedata import check_framedata
from .py_utils import check_iterable
from .logger import logger

__all__ = ['list_fits_files', 'FitsFileGroup']


def list_fits_files(location, fits_extensions=None,
                    glob_include=None, glob_exclude=None):
    """List all fist files in a directory, if compressed or not."""
    if fits_extensions is None:
        fits_extensions = _fits_extensions

    if not check_iterable(fits_extensions):
        fits_extensions = [fits_extensions]

    f = []
    for i in fits_extensions:
        files = glob.glob(os.path.join(location, '**/*'+i), recursive=True)
        # filter only glob include
        if glob_include is not None:
            if not check_iterable(glob_include):
                glob_include = [glob_include]
            for inc in glob_include:
                files = [i for i in files if fnmatch.fnmatch(i, inc)]
        f.extend(files)

    # Filter excluded files
    if glob_exclude is not None:
        if not check_iterable(glob_exclude):
            glob_exclude = [glob_exclude]
        for exc in glob_exclude:
            f = [i for i in f if not fnmatch.fnmatch(i, exc)]

    files = sorted(f)
    # files = [os.path.join(location, i) for i in files]
    return files


_headers = 'headers'
_metadata = 'astropop_metadata'
_files_col = '__file'
_keycolstable = 'astropop_keyword2column'
_keywords_col = 'keyword'
_columns_col = 'column'


class FitsFileGroup():
    """Easy handle groups of fits files."""

    def __init__(self, location=None, files=None, ext=0,
                 compression=False, database=':memory:', **kwargs):
        self._ext = ext
        self._extensions = kwargs.get('fits_ext', None)
        self._include = kwargs.get('glob_include')
        self._exclude = kwargs.get('glob_exclude')

        self._db = SQLDatabase(database)
        if database == ':memory:':
            self._db_dir = None
        else:
            self._db_dir = Path(database).resolve().parent

        self._read_db(files, location, compression, kwargs.get('update', 0))

    def _list_files(self, files, location, compression):
        extensions = self._extensions
        if extensions is None:
            if compression:
                extensions = _fits_extensions_with_compress
            else:
                extensions = _fits_extensions

        if files is not None and location is not None:
            raise ValueError('You can only specify either files or location.')
        if files is None and location is not None:
            files = list_fits_files(location, extensions,
                                    self._include, self._exclude)
        if files is None:
            files = []
        return files

    def _read_db(self, files, location, compression, update=False):
        """Read the database and generate the summary if needed."""
        initialized = _metadata in self._db.table_names
        if location is not None:
            location = str(location)

        if not initialized:
            self._db.add_table(_metadata)
            self._db.add_rows(_metadata, {'GLOB_INCLUDE': self._include,
                                          'GLOB_EXCLUDE': self._exclude,
                                          'LOCATION': location,
                                          'COMPRESSION': compression,
                                          'EXT': self._ext},
                              add_columns=True)
            self._db.add_column(_metadata, 'FITS_EXT', self._extensions)
            self._db.add_table(_keycolstable)
            self._db.add_column(_keycolstable, _keywords_col)
            self._db.add_column(_keycolstable, _columns_col)

        self._include = self._db[_metadata, 'glob_include'][0]
        self._exclude = self._db[_metadata, 'glob_exclude'][0]
        self._extensions = self._db[_metadata, 'fits_ext'].values
        if self._extensions == [None]:
            self._extensions = None
        self._ext = self._db[_metadata, 'ext'][0]
        self._location = self._db[_metadata, 'location'][0]
        self._compression = self._db[_metadata, 'compression'][0]

        cmap = SQLColumnMap(self._db, _keycolstable,
                            _keywords_col, _columns_col)
        self._table = SQLTable(self._db, _headers, colmap=cmap)

        if update or not initialized:
            self.update(files, location, compression)

    @property
    def files(self):
        """List files in the group."""
        return [self.full_path(i) for i in range(len(self))]

    @property
    def summary(self):
        """Get a readonly table with summary of the fits files."""
        return self._table.as_table()

    def __copy__(self, indexes=None):
        """Copy the current instance to a new object."""
        if indexes is None:
            db = self._db.copy()
        else:
            db = self._db.copy(indexes={_headers: indexes})

        nfg = object.__new__(FitsFileGroup)
        nfg._db = db
        nfg._db_dir = self._db_dir  # need to keep the db_dir for the iterators
        nfg._read_db(None, None, None, False)
        return nfg

    def __len__(self):
        """Get the number of files in the group."""
        return len(self._table)

    def filtered(self, keywords):
        """Create a new FitsFileGroup with only filtered files."""
        try:
            indexes = self._table.select(columns=[_ID_KEY],
                                         where=keywords)
        except sql.OperationalError:
            indexes = []
        if len(indexes) == 0:
            return self.__copy__(indexes=[])
        indexes = np.array(indexes).ravel() - 1
        return self.__copy__(indexes)

    def group_by(self, keywords):
        """Create FitsFileGroups grouped by keywords."""
        summary = self.summary
        id_key = 'id'
        while id_key in summary.colnames:
            id_key += '_'
        summary.add_column(Column(np.arange(len(summary))),
                           name=id_key)

        grouped = summary.group_by(keywords)
        for g in grouped.groups:
            yield self.__copy__(indexes=list(g[id_key]))

    def update(self, files=None, location=None, compression=False):
        """Update the database with the current files."""
        if _headers in self._db.table_names:
            self._db.drop_table(_headers)

        self._db.add_table(_headers)
        location = location or self._location
        compression = compression or self._compression
        files = self._list_files(files, location, compression)
        for i, f in enumerate(files):
            logger.info('reading file %i from %i: %s', i, len(files), f)
            try:
                self.add_file(f)
            except Exception as e:
                # just log error instead of raise an exception
                logger.error('Error reading file %s: %s', f, e)

    def values(self, keyword, unique=False):
        """Return the values of a keyword in the summary.

        If unique, only unique values returned.
        """
        vals = self._table[keyword].values
        if unique:
            vals = list(set(vals))
        return vals

    def add_column(self, name, values=None):
        """Add a new column to the summary."""
        self._table.add_column(name, data=values)

    def add_file(self, file):
        """Add a new file to the group."""
        header = fits.open(file)[self._ext].header
        logger.debug('reading file %s', file)
        file = self.relative_path(file)
        hdr = {_files_col: file}
        hdr.update(dict(header))
        hdr.pop('COMMENT', None)
        hdr.pop('HISTORY', None)
        hdr.pop('', None)
        self._table.add_rows(hdr, add_columns=True)

    def full_path(self, file):
        """Get the full path of a file in the group.

        Parameters
        ----------
        file : str or int
            If string, the file name. If int, the index of the file.

        Returns
        -------
        path : str
            Full path of the file.
        """
        if isinstance(file, int):
            file = self._table[_files_col][file]
        if self._db_dir is not None:
            return os.path.join(self._db_dir, file)
        return file

    def relative_path(self, file):
        """Get the relative path of a file.

        Parameters
        ----------
        file : str
            Full path of the file.

        Returns
        -------
        path : str
            Relative path of the file.
        """
        if self._db_dir is not None:
            return os.path.relpath(file, self._db_dir)
        return file

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._table[item].values

        # returning FitsFileGroups
        if isinstance(item, (int, np.integer)):
            # single index will be interpreted as a single file group
            return self.__copy__(indexes=[item])
        if (isinstance(item, slice)):
            item = list(range(*item.indices(len(self))))
        if isinstance(item, (np.ndarray, list)):
            item = np.array(item)
            return self.__copy__(indexes=item)

        raise KeyError(f'{item}')

    def __setitem__(self, item, value):
        """Set the value of a keyword in the summary."""
        self._table[item] = value

    def _intern_yelder(self, ext=None, ret_type=None, **kwargs):
        """Iterate over files."""
        ext = ext if ext is not None else self._ext
        for i in self.files:
            if self._db_dir is not None:
                i = os.path.join(self._db_dir, i)

            if ret_type == 'header':
                yield fits.open(i, **kwargs)[ext].header
            if ret_type == 'data':
                yield fits.open(i, **kwargs)[ext].data
            if ret_type == 'hdu':
                yield fits.open(i, **kwargs)[ext]
            if ret_type == 'framedata':
                yield check_framedata(i, hdu=ext, **kwargs)

    def hdus(self, ext=None, **kwargs):
        """Read the files and iterate over their HDUs."""
        return self._intern_yelder(ext=ext, ret_type='hdu', **kwargs)

    def headers(self, ext=None, **kwargs):
        """Read the files and iterate over their headers."""
        return self._intern_yelder(ext=ext, ret_type='header', **kwargs)

    def data(self, ext=None, **kwargs):
        """Read the files and iterate over their data."""
        return self._intern_yelder(ext=ext, ret_type='data', **kwargs)

    def framedata(self, ext=None, **kwargs):
        """Read the files and iterate over their data."""
        return self._intern_yelder(ext=ext, ret_type='framedata', **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import numpy as np
import os
import pytest
from urllib import request
from astroquery.skyview import SkyView
from astropy.coordinates import Angle, SkyCoord
from astropy.config import get_cache_dir
from astropy.io import fits
from astropy.nddata.ccddata import _generate_wcs_and_update_header
from astropy.wcs import WCS

from astropop.astrometry.astrometrynet import _solve_field, fit_wcs, \
                                              solve_astrometry_image, \
                                              solve_astrometry_xy, \
                                              solve_astrometry_hdu
from astropop.astrometry.manual_wcs import wcs_from_coords
from astropop.astrometry.coords_utils import guess_coordinates
from astropop.photometry.aperture import aperture_photometry
from astropop.photometry.detection import starfind

from astropop.testing import *


def get_image_index():
    cache = get_cache_dir()
    ast_data = os.path.dirname(_solve_field)
    ast_data = os.path.dirname(ast_data)
    ast_data = os.path.join(ast_data, 'data')
    os.makedirs(ast_data, exist_ok=True)
    index = 'index-4107.fits'  # index-4202-28.fits'
    d = 'http://broiler.astrometry.net/~dstn/4100/' + index
    f = os.path.join(ast_data, index)
    if not os.path.isfile(f):
        request.urlretrieve(d, f)  # nosec
    name = os.path.join(cache, 'm20_dss.fits')
    if not os.path.isfile(name):
        s = SkyView.get_images('M20', radius=Angle('60arcmin'),
                               pixels=2048, survey='DSS')
        s[0][0].writeto(name)
    return name, f


def compare_wcs(wcs, nwcs):
    for i in [(100, 100), (1000, 1500), (357.5, 948.2), (2015.1, 403.7)]:
        res1 = np.array(wcs.all_pix2world(*i, 0))
        res2 = np.array(nwcs.all_pix2world(*i, 0))
        assert_almost_equal(res1, res2, decimal=3)


skip_astrometry = pytest.mark.skipif("_solve_field is None or "
                                     "os.getenv('SKIP_TEST_ASTROMETRY', "
                                     "False)")


@skip_astrometry
def test_solve_astrometry_hdu(tmpdir):
    data, index = get_image_index()
    hdu = fits.open(data)[0]
    header, wcs = _generate_wcs_and_update_header(hdu.header)
    hdu.header = header
    nwcs = solve_astrometry_hdu(hdu, return_wcs=True)
    assert_true(isinstance(nwcs, WCS))
    assert_equal(nwcs.naxis, 2)
    compare_wcs(wcs, nwcs)


@skip_astrometry
def test_solve_astrometry_xyl(tmpdir):
    data, index = get_image_index()
    hdu = fits.open(data)[0]
    header, wcs = _generate_wcs_and_update_header(hdu.header)
    hdu.header = header
    sources = starfind(hdu.data, 10, np.median(hdu.data),
                       np.std(hdu.data), 4)
    phot = aperture_photometry(hdu.data, sources['x'], sources['y'])
    imw, imh = hdu.data.shape
    nwcs = solve_astrometry_xy(phot['x'], phot['y'], phot['flux'], header,
                               imw, imh, return_wcs=True)
    assert_is_instance(nwcs, WCS)
    assert_equal(nwcs.naxis, 2)
    compare_wcs(wcs, nwcs)


@skip_astrometry
def test_solve_astrometry_image(tmpdir):
    data, index = get_image_index()
    hdu = fits.open(data)[0]
    header, wcs = _generate_wcs_and_update_header(hdu.header)
    hdu.header = header
    name = tmpdir.join('testimage.fits').strpath
    hdu.writeto(name)
    nwcs = solve_astrometry_image(name, return_wcs=True)
    assert_is_instance(nwcs, WCS)
    assert_equal(nwcs.naxis, 2)
    compare_wcs(wcs, nwcs)


@skip_astrometry
def test_fit_wcs(tmpdir):
    data, index = get_image_index()
    hdu = fits.open(data)[0]
    imw, imh = hdu.data.shape
    header, wcs = _generate_wcs_and_update_header(hdu.header)
    hdu.header = header
    sources = starfind(hdu.data, 10, np.median(hdu.data),
                       np.std(hdu.data), 4)
    sources['ra'], sources['dec'] = wcs.all_pix2world(sources['x'],
                                                      sources['y'], 1)
    nwcs = fit_wcs(sources['x'], sources['y'], sources['ra'], sources['dec'],
                   imw, imh)
    assert_is_instance(nwcs, WCS)
    assert_equal(nwcs.naxis, 2)
    compare_wcs(wcs, nwcs)


def test_manual_wcs_top():
    # Checked with DS9
    x, y = (11, 11)
    ra, dec = (10.0, 0.0)
    ps = 36  # arcsec/px
    north = 'top'  # north to right
    wcs = wcs_from_coords(x, y, ra, dec, ps, north)
    assert_almost_equal(wcs.all_pix2world(11, 11, 1), (ra, dec))
    assert_almost_equal(wcs.all_pix2world(11, 21, 1), (10.0, 0.1))
    assert_almost_equal(wcs.all_pix2world(21, 11, 1), (9.9, 0.0))
    assert_almost_equal(wcs.all_pix2world(1, 11, 1), (10.1, 0.0))
    assert_almost_equal(wcs.all_pix2world(11, 1, 1), (10.0, -0.1))
    assert_almost_equal(wcs.all_pix2world(21, 21, 1), (9.9, 0.1))
    assert_almost_equal(wcs.all_pix2world(1, 21, 1), (10.1, 0.1))
    assert_almost_equal(wcs.all_pix2world(1, 1, 1), (10.1, -0.1))


def test_manual_wcs_left():
    # Checked with DS9
    x, y = (11, 11)
    ra, dec = (10.0, 0.0)
    ps = 36  # arcsec/px
    north = 'left'  # north to right
    wcs = wcs_from_coords(x, y, ra, dec, ps, north)
    assert_almost_equal(wcs.all_pix2world(11, 11, 1), (ra, dec))
    assert_almost_equal(wcs.all_pix2world(11, 21, 1), (9.9, 0.0))
    assert_almost_equal(wcs.all_pix2world(21, 11, 1), (10.0, -0.1))
    assert_almost_equal(wcs.all_pix2world(1, 11, 1), (10.0, 0.1))
    assert_almost_equal(wcs.all_pix2world(11, 1, 1), (10.1, 0.0))
    assert_almost_equal(wcs.all_pix2world(21, 21, 1), (9.9, -0.1))
    assert_almost_equal(wcs.all_pix2world(1, 21, 1), (9.9, 0.1))
    assert_almost_equal(wcs.all_pix2world(1, 1, 1), (10.1, 0.1))


def test_manual_wcs_bottom():
    # Checked with DS9
    x, y = (11, 11)
    ra, dec = (10.0, 0.0)
    ps = 36  # arcsec/px
    north = 'bottom'  # north to right
    wcs = wcs_from_coords(x, y, ra, dec, ps, north)
    assert_almost_equal(wcs.all_pix2world(11, 11, 1), (ra, dec))
    assert_almost_equal(wcs.all_pix2world(11, 21, 1), (10.0, -0.1))
    assert_almost_equal(wcs.all_pix2world(21, 11, 1), (10.1, 0.0))
    assert_almost_equal(wcs.all_pix2world(1, 11, 1), (9.9, 0.0))
    assert_almost_equal(wcs.all_pix2world(11, 1, 1), (10.0, 0.1))
    assert_almost_equal(wcs.all_pix2world(21, 21, 1), (10.1, -0.1))
    assert_almost_equal(wcs.all_pix2world(1, 21, 1), (9.9, -0.1))
    assert_almost_equal(wcs.all_pix2world(1, 1, 1), (9.9, 0.1))


def test_manual_wcs_right():
    # Checked with DS9
    x, y = (11, 11)
    ra, dec = (10.0, 0.0)
    ps = 36  # arcsec/px
    north = 'right'
    wcs = wcs_from_coords(x, y, ra, dec, ps, north)
    assert_almost_equal(wcs.all_pix2world(11, 11, 1), (ra, dec))
    assert_almost_equal(wcs.all_pix2world(11, 21, 1), (10.1, 0.0))
    assert_almost_equal(wcs.all_pix2world(21, 11, 1), (10.0, 0.1))
    assert_almost_equal(wcs.all_pix2world(1, 11, 1), (10.0, -0.1))
    assert_almost_equal(wcs.all_pix2world(11, 1, 1), (9.9, 0.))
    assert_almost_equal(wcs.all_pix2world(21, 21, 1), (10.1, 0.1))
    assert_almost_equal(wcs.all_pix2world(1, 21, 1), (10.1, -0.1))
    assert_almost_equal(wcs.all_pix2world(1, 1, 1), (9.9, -0.1))


def test_manual_wcs_angle():
    # Checked with DS9
    x, y = (11, 11)
    ra, dec = (10.0, 0.0)
    ps = 36  # arcsec/px
    north = 45
    wcs = wcs_from_coords(x, y, ra, dec, ps, north)
    assert_almost_equal(wcs.all_pix2world(11, 11, 1), (ra, dec))
    assert_almost_equal(wcs.all_pix2world(21, 21, 1), (10.0, 0.14142))
    assert_almost_equal(wcs.all_pix2world(1, 1, 1), (10.0, -0.14142))
    assert_almost_equal(wcs.all_pix2world(1, 21, 1), (10.14142, 0.0))
    assert_almost_equal(wcs.all_pix2world(21, 1, 1), (9.858579, 0.0))


def test_manual_wcs_top_flip_ra():
    # Checked with DS9
    x, y = (11, 11)
    ra, dec = (10.0, 0.0)
    ps = 36  # arcsec/px
    north = 'top'
    flip = 'ra'
    wcs = wcs_from_coords(x, y, ra, dec, ps, north, flip=flip)
    assert_almost_equal(wcs.all_pix2world(11, 11, 1), (ra, dec))
    assert_almost_equal(wcs.all_pix2world(11, 21, 1), (10.0, 0.1))
    assert_almost_equal(wcs.all_pix2world(21, 11, 1), (10.1, 0.0))
    assert_almost_equal(wcs.all_pix2world(1, 11, 1), (9.9, 0.0))
    assert_almost_equal(wcs.all_pix2world(11, 1, 1), (10.0, -0.1))
    assert_almost_equal(wcs.all_pix2world(21, 21, 1), (10.1, 0.1))
    assert_almost_equal(wcs.all_pix2world(1, 21, 1), (9.9, 0.1))
    assert_almost_equal(wcs.all_pix2world(1, 1, 1), (9.9, -0.1))


def test_manual_wcs_top_flip_dec():
    # Checked with DS9
    x, y = (11, 11)
    ra, dec = (10.0, 0.0)
    ps = 36  # arcsec/px
    north = 'top'
    flip = 'dec'
    wcs = wcs_from_coords(x, y, ra, dec, ps, north, flip=flip)
    assert_almost_equal(wcs.all_pix2world(11, 11, 1), (ra, dec))
    assert_almost_equal(wcs.all_pix2world(11, 21, 1), (10.0, -0.1))
    assert_almost_equal(wcs.all_pix2world(21, 11, 1), (9.9, 0.0))
    assert_almost_equal(wcs.all_pix2world(1, 11, 1), (10.1, 0.0))
    assert_almost_equal(wcs.all_pix2world(11, 1, 1), (10.0, 0.1))
    assert_almost_equal(wcs.all_pix2world(21, 21, 1), (9.9, -0.1))
    assert_almost_equal(wcs.all_pix2world(1, 21, 1), (10.1, -0.1))
    assert_almost_equal(wcs.all_pix2world(1, 1, 1), (10.1, 0.1))


def test_manual_wcs_top_flip_all():
    # Checked with DS9
    x, y = (11, 11)
    ra, dec = (10.0, 0.0)
    ps = 36  # arcsec/px
    north = 'top'
    flip = 'all'
    wcs = wcs_from_coords(x, y, ra, dec, ps, north, flip=flip)
    assert_almost_equal(wcs.all_pix2world(11, 21, 1), (10.0, -0.1))
    assert_almost_equal(wcs.all_pix2world(21, 11, 1), (10.1, 0.0))
    assert_almost_equal(wcs.all_pix2world(1, 11, 1), (9.9, 0.0))
    assert_almost_equal(wcs.all_pix2world(11, 1, 1), (10.0, 0.1))
    assert_almost_equal(wcs.all_pix2world(21, 21, 1), (10.1, -0.1))
    assert_almost_equal(wcs.all_pix2world(1, 21, 1), (9.9, -0.1))
    assert_almost_equal(wcs.all_pix2world(1, 1, 1), (9.9, 0.1))


def test_raise_north_angle():
    with pytest.raises(ValueError) as exc:
        wcs_from_coords(0, 0, 0, 0, 0, 'not a direction')
        assert_in('invalid value for north', str(exc.value))


def test_guess_coords_float():
    ra = 10.0
    dec = 0.0
    assert_equal(guess_coordinates(ra, dec, skycoord=False), (ra, dec))


def test_guess_coords_strfloat():
    ra = "10.0"
    dec = "-27.0"
    assert_equal(guess_coordinates(ra, dec, skycoord=False), (10, -27))


def test_guess_coords_hexa_space():
    ra = "1 00 00"
    dec = "-43 30 00"
    assert_almost_equal(guess_coordinates(ra, dec, skycoord=False),
                              (15.0, -43.5))


def test_guess_coords_hexa_dots():
    ra = "1:00:00"
    dec = "-43:30:00"
    assert_almost_equal(guess_coordinates(ra, dec, skycoord=False),
                              (15.0, -43.5))


def test_guess_coords_skycord_float():
    ra = 10.0
    dec = 0.0
    sk = guess_coordinates(ra, dec, skycoord=True)
    assert_is_instance(sk, SkyCoord)
    assert_equal(sk.ra.degree, ra)
    assert_equal(sk.dec.degree, dec)


def test_guess_coords_skycord_hexa():
    ra = "1:00:00"
    dec = "00:00:00"
    sk = guess_coordinates(ra, dec, skycoord=True)
    assert_is_instance(sk, SkyCoord)
    assert_true(sk.ra.degree - 15 < 1e-8)
    assert_true(sk.dec.degree - 0 < 1e-8)


def test_guess_coords_list_hexa():
    ra = ["1:00:00", "2:30:00"]
    dec = ["00:00:00", "1:00:00"]
    sra, sdec = guess_coordinates(ra, dec, skycoord=False)
    assert_almost_equal(sra, [15, 37.5])
    assert_almost_equal(sdec, [0, 1])


def test_guess_coords_list_float():
    ra = [10.0, 15, 20]
    dec = [0.0, 1.0, -1.0]
    sra, sdec = guess_coordinates(ra, dec, skycoord=False)
    assert_equal(sra, ra)
    assert_equal(sdec, dec)


def test_guess_coords_list_diff():
    ra = np.arange(10)
    dec = np.arange(15)
    with pytest.raises(ValueError):
        guess_coordinates(ra, dec)


def test_guess_coords_list_nolist():
    ra = np.arange(10)
    dec = 1
    with pytest.raises(ValueError):
        guess_coordinates(ra, dec)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
import numpy as np
from astropop.image.register import AsterismRegister, \
                                    CrossCorrelationRegister, \
                                    register_framedata_list, \
                                    compute_shift_list
from astropop.framedata import FrameData
from astropop.testing import *

from .test_detection import gen_position_flux, gen_image


def gen_positions_transformed(x, y, flux, dx, dy, limits,
                              rotation=None, rotation_center=None):
    """Generate translated positions."""
    x, y = x+dx, y+dy

    if rotation is not None:
        rotation_center = rotation_center or np.array(limits)/2
        cx, cy = rotation_center
        theta = np.deg2rad(rotation)
        nx = cx + (x-cx)*np.cos(theta) - (y-cy)*np.sin(theta)
        ny = cy + (x-cx)*np.sin(theta) + (y-cy)*np.cos(theta)
        x, y = nx, ny

    # ensure all positions are inside the image
    mask = x >= 0
    mask &= x <= limits[0]
    mask &= y >= 0
    mask &= y <= limits[1]
    where = np.where(mask)

    return x[where], y[where], flux[where]


class Test_AsterismRegister:
    @pytest.mark.parametrize('shift', [(25, 32), (-12, 5), (23.42, 12.43)])
    def test_compute_transform_translation(self, shift):
        size = (1024, 1024)
        sky = 800
        n = 100
        rdnoise = 5
        x, y, f = gen_position_flux(np.array(size)+80, n, 1e4, 4e6)
        x -= 40
        y -= 40
        sx, sy = shift

        x1, y1, flux1 = gen_positions_transformed(x, y, f, 0, 0, size)
        im1 = gen_image(size, x1, y1, flux1,
                        sky, rdnoise, sigma=2)

        x2, y2, flux2 = gen_positions_transformed(x, y, f, sx, sy, size)
        im2 = gen_image(size, x2, y2, flux2,
                        sky, rdnoise, sigma=2)

        ar = AsterismRegister()
        tform = ar._compute_transform(im1, im2)

        assert_almost_equal(tform.translation, shift, decimal=1)
        assert_almost_equal(tform.rotation, 0, decimal=3)
        assert_almost_equal(tform.scale, 1, decimal=4)

    def test_compute_transform_rotation(self):
        size = (1024, 1024)
        sky = 800
        n = 300
        rdnoise = 5
        x, y, f = gen_position_flux(2*np.array(size), n, 1e4, 4e6)
        x -= 500
        y -= 500

        x1, y1, flux1 = gen_positions_transformed(x, y, f, 0, 0, size)
        im1 = gen_image(size, x1, y1, flux1,
                        sky, rdnoise, sigma=2)

        x2, y2, flux2 = gen_positions_transformed(x, y, f, 0, 0, size,
                                                  rotation=35.2)
        im2 = gen_image(size, x2, y2, flux2,
                        sky, rdnoise, sigma=2)

        ar = AsterismRegister()
        tform = ar._compute_transform(im1, im2)

        assert_almost_equal(tform.rotation, np.deg2rad(35.2), decimal=3)
        # the translation is needed due to the form skimage handles rotation
        assert_almost_equal(tform.translation, [388.7, -201.5], decimal=0)
        assert_almost_equal(tform.scale, 1, decimal=4)


class Test_CrossCorrelationRegister:
    @pytest.mark.parametrize('shift', [(25, 32), (-12, 5), (23.42, 12.43)])
    def test_compute_transform(self, shift):
        size = (1024, 1024)
        sky = 800
        n = 60
        rdnoise = 5
        x, y, f = gen_position_flux(np.array(size)+80, n, 1e4, 4e6)
        x -= 40
        y -= 40
        sx, sy = shift

        x1, y1, flux1 = gen_positions_transformed(x, y, f, 0, 0, size)
        im1 = gen_image(size, x1, y1, flux1,
                        sky, rdnoise, sigma=2)

        x2, y2, flux2 = gen_positions_transformed(x, y, f, sx, sy, size)
        im2 = gen_image(size, x2, y2, flux2,
                        sky, rdnoise, sigma=2)

        ccr = CrossCorrelationRegister(upsample_factor=10)
        tform = ccr._compute_transform(im1, im2)

        assert_almost_equal(tform.translation, shift, decimal=1)
        assert_almost_equal(tform.rotation, 0, decimal=3)
        assert_almost_equal(tform.scale, 1, decimal=4)


class Test_Registration:
    @pytest.mark.parametrize('cval,fill', [(0, 0),
                                           ('median', 1.0),
                                           ('mean', 1.51)])
    def test_register_image(self, cval, fill):
        im1 = [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 2, 2, 2, 1, 1, 1, 1, 1],
               [1, 2, 4, 6, 4, 2, 1, 1, 1, 1],
               [1, 2, 6, 8, 6, 2, 1, 1, 1, 1],
               [1, 2, 4, 6, 4, 2, 1, 1, 1, 1],
               [1, 1, 2, 2, 2, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]

        im2 = [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 2, 2, 2, 1, 1, 1],
               [1, 1, 1, 2, 4, 6, 4, 2, 1, 1],
               [1, 1, 1, 2, 6, 8, 6, 2, 1, 1],
               [1, 1, 1, 2, 4, 6, 4, 2, 1, 1],
               [1, 1, 1, 1, 2, 2, 2, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]

        expect = np.array(im1, dtype='f8')
        expect[0, :] = fill
        expect[:, -2:] = fill

        mask = np.zeros_like(im2, dtype=bool)
        mask[0, :] = 1
        mask[:, -2:] = 1

        ar = CrossCorrelationRegister()
        im_reg, mask_reg, tform = ar.register_image(np.array(im1), np.array(im2),
                                                    cval=cval)

        assert_equal(im_reg, expect)
        assert_equal(mask_reg, mask)
        assert_equal(tform.translation, [2, -1])

    @pytest.mark.parametrize('inplace', [True, False])
    @pytest.mark.parametrize('cval,fill', [(0, 0),
                                           ('median', 1.0),
                                           ('mean', 1.51)])
    def test_register_framedata(self, inplace, cval, fill):
        im1 = [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 2, 2, 2, 1, 1, 1, 1, 1],
               [1, 2, 4, 6, 4, 2, 1, 1, 1, 1],
               [1, 2, 6, 8, 6, 2, 1, 1, 1, 1],
               [1, 2, 4, 6, 4, 2, 1, 1, 1, 1],
               [1, 1, 2, 2, 2, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]

        im2 = [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 2, 2, 2, 1, 1, 1],
               [1, 1, 1, 2, 4, 6, 4, 2, 1, 1],
               [1, 1, 1, 2, 6, 8, 6, 2, 1, 1],
               [1, 1, 1, 2, 4, 6, 4, 2, 1, 1],
               [1, 1, 1, 1, 2, 2, 2, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]

        expect = np.array(im1, dtype='f8')
        expect[0, :] = fill
        expect[:, -2:] = fill

        mask = np.zeros_like(im2, dtype=bool)
        mask[0, :] = 1
        mask[:, -2:] = 1

        expect_unct = np.ones_like(im2, dtype='f8')
        expect_unct[0, :] = np.nan
        expect_unct[:, -2:] = np.nan

        frame1 = FrameData(im1, dtype='f8')
        frame1.meta['moving'] = False
        frame1.uncertainty = np.ones_like(im1)
        frame2 = FrameData(im2, dtype='f8')
        frame2.meta['moving'] = True
        frame2.uncertainty = np.ones_like(im2)

        ar = CrossCorrelationRegister()
        frame_reg = ar.register_framedata(frame1, frame2,
                                          cval=cval, inplace=inplace)

        assert_equal(frame_reg.data, expect)
        assert_equal(frame_reg.mask, mask)
        assert_equal(frame_reg.uncertainty, expect_unct)
        assert_equal(frame_reg.meta['astropop registration'],
                     'cross-correlation')
        assert_equal(frame_reg.meta['astropop registration_shift_x'], 2)
        assert_equal(frame_reg.meta['astropop registration_shift_y'], -1)
        assert_equal(frame_reg.meta['astropop registration_rot'], 0)
        assert_equal(frame_reg.meta['moving'], True)
        if inplace:
            assert_is(frame_reg, frame2)
        else:
            assert_is_not(frame_reg, frame2)

    def test_register_image_equal(self):
        im = gen_image((50, 50), [25], [25], [10000], 10, 0, sigma=3)
        ar = CrossCorrelationRegister()
        im_reg, mask_reg, tform = ar.register_image(im, im)
        assert_is(im_reg, im)
        assert_equal(im_reg, im)
        assert_equal(mask_reg, np.zeros_like(im))
        assert_equal(tform.translation, [0, 0])

    @pytest.mark.parametrize('inplace', [True, False])
    def test_register_frame_equal(self, inplace):
        im = gen_image((50, 50), [25], [25], [10000], 10, 0, sigma=3)
        im = FrameData(im)
        ar = CrossCorrelationRegister()
        im_reg = ar.register_framedata(im, im, inplace=inplace)
        if inplace:
            assert_is(im_reg, im)
        else:
            assert_is_not(im_reg, im)
        assert_equal(im_reg.data, im.data)
        assert_equal(im_reg.mask, np.zeros_like(im))
        assert_equal(im_reg.meta['astropop registration_shift_x'], 0)
        assert_equal(im_reg.meta['astropop registration_shift_y'], 0)


class Test_Register_FrameData_List:
    _shifts = [(0, 0), (-1, 22.4), (15.5, 3.2), (2.2, -1.75), (-5.4, 0.5)]

    def gen_frame_list(self, size):
        sky = 800
        rdnoise = 10
        n = 100
        x, y, f = gen_position_flux(np.array(size)+80, n, 1e4, 4e6)
        x -= 40
        y -= 40

        frame_list = []
        for shift in self._shifts:
            x1, y1, flux1 = gen_positions_transformed(x, y, f, *shift, size)
            im1 = gen_image(size, x1, y1, flux1,
                            sky, rdnoise, sigma=2)
            frame = FrameData(im1, meta={'test expect_shift_x': shift[0],
                                         'test expect_shift_y': shift[1]})
            frame_list.append(frame)

        return frame_list

    def test_error_unkown_algorithm(self):
        with pytest.raises(ValueError, match='Algorithm noexisting unknown.'):
            register_framedata_list([FrameData(None) for i in range(10)],
                                    algorithm='noexisting')
        with pytest.raises(ValueError, match='Algorithm noexisting unknown.'):
            compute_shift_list([FrameData(None) for i in range(10)],
                               algorithm='noexisting')

    def test_error_non_framedata(self):
        with pytest.raises(TypeError, match='Only a list of FrameData'):
            register_framedata_list([np.zeros((10, 10)) for i in range(10)])
        with pytest.raises(TypeError, match='Only a list of FrameData'):
            compute_shift_list([np.zeros((10, 10)) for i in range(10)])

    def test_error_non_iterable_list(self):
        with pytest.raises(TypeError):
            register_framedata_list(10)
        with pytest.raises(TypeError):
            compute_shift_list(10)

    def test_error_incompatible_shapes(self):
        frame_list = [FrameData(np.zeros((i+1, i+1))) for i in range(10)]
        with pytest.raises(ValueError, match='incompatible shapes'):
            register_framedata_list(frame_list)
        with pytest.raises(ValueError, match='incompatible shapes'):
            compute_shift_list(frame_list)

    @pytest.mark.parametrize('inplace', [True, False])
    def test_register_framedata_crosscorr(self, inplace):
        frame_list = self.gen_frame_list((1024, 1024))
        reg_list = register_framedata_list(frame_list,
                                           algorithm='cross-correlation',
                                           inplace=inplace,
                                           upsample_factor=10, space='real')
        assert_equal(len(frame_list), len(reg_list))
        for org, reg in zip(frame_list, reg_list):
            if inplace:
                assert_is(org, reg)
            else:
                assert_is_not(org, reg)
            for i in ['x', 'y']:
                ap_reg_shift = reg.meta[f'astropop registration_shift_{i}']
                ex_reg_shift = org.meta[f'test expect_shift_{i}']
                assert_almost_equal(ap_reg_shift, ex_reg_shift, decimal=0)

        if not inplace:
            shift_list = compute_shift_list(frame_list,
                                            algorithm='cross-correlation',
                                            upsample_factor=10, space='real')
            assert_equal(len(frame_list), len(shift_list))
            assert_almost_equal(shift_list, self._shifts, decimal=0)

    @pytest.mark.parametrize('inplace', [True, False])
    def test_register_framedata_asterism(self, inplace):
        frame_list = self.gen_frame_list((1024, 1024))
        reg_list = register_framedata_list(frame_list,
                                           algorithm='asterism-matching',
                                           inplace=inplace,
                                           max_control_points=30,
                                           detection_threshold=5)
        assert_equal(len(frame_list), len(reg_list))
        for org, reg in zip(frame_list, reg_list):
            if inplace:
                assert_is(org, reg)
            else:
                assert_is_not(org, reg)
            for i in ['x', 'y']:
                ap_reg_shift = reg.meta[f'astropop registration_shift_{i}']
                ex_reg_shift = org.meta[f'test expect_shift_{i}']
                assert_almost_equal(ap_reg_shift, ex_reg_shift, decimal=0)

        if not inplace:
            shift_list = compute_shift_list(frame_list,
                                            algorithm='asterism-matching',
                                            max_control_points=30,
                                            detection_threshold=5)
            assert_equal(len(frame_list), len(shift_list))
            assert_almost_equal(shift_list, self._shifts, decimal=0)

    def test_register_framedata_list_ref_image(self):
        frame_list = self.gen_frame_list((1024, 1024))
        reg_list = register_framedata_list(frame_list,
                                           algorithm='asterism-matching',
                                           ref_image=4,
                                           max_control_points=30,
                                           detection_threshold=5)
        assert_equal(len(frame_list), len(reg_list))
        ref_shift = np.array(self._shifts[4])

        for org, reg in zip(frame_list, reg_list):
            for i, ref in zip(['x', 'y'], ref_shift):
                ap_reg_shift = reg.meta[f'astropop registration_shift_{i}']
                ex_reg_shift = org.meta[f'test expect_shift_{i}'] - ref
                assert_almost_equal(ap_reg_shift, ex_reg_shift, decimal=0)

        shift_list = compute_shift_list(frame_list,
                                        ref_image=4,
                                        algorithm='asterism-matching',
                                        max_control_points=30,
                                        detection_threshold=5)
        assert_equal(len(frame_list), len(shift_list))
        assert_almost_equal(shift_list, self._shifts - ref_shift, decimal=0)

    def test_register_framedata_list_clip(self):
        frame_list = self.gen_frame_list((512, 1024))
        reg_list = register_framedata_list(frame_list, clip_output=True,
                                           inplace=True,
                                           algorithm='asterism-matching',
                                           max_control_points=30,
                                           detection_threshold=5)

        assert_equal(len(frame_list), len(reg_list))
        for org, reg in zip(frame_list, reg_list):
            assert_is(org, reg)
            for i in ['x', 'y']:
                ap_reg_shift = reg.meta[f'astropop registration_shift_{i}']
                ex_reg_shift = org.meta[f'test expect_shift_{i}']
                assert_almost_equal(ap_reg_shift, ex_reg_shift, decimal=0)
            assert_equal(reg.meta['astropop trimmed_section'], '6:-16,2:-23')
            # x: 6:-16, y: 2:-23
            assert_equal(reg.shape, (1024-23-2, 512-6-16))
            # no masked pixel should remain
            assert_false(np.any(reg.mask))

    def test_register_framedata_list_skip_failure_default(self):
        # defult behavior is raise error
        frame_list = self.gen_frame_list((512, 1024))
        frame_list[2].data = np.ones((1024, 512))

        with pytest.raises(TypeError):
            register_framedata_list(frame_list, algorithm='asterism-matching')

        with pytest.raises(TypeError):
            compute_shift_list(frame_list, algorithm='asterism-matching')

    def test_register_framedata_list_skip_failure_false(self):
        frame_list = self.gen_frame_list((512, 1024))
        frame_list[2].data = np.ones((1024, 512))

        with pytest.raises(TypeError):
            register_framedata_list(frame_list, algorithm='asterism-matching',
                                    skip_failure=False)

        with pytest.raises(TypeError):
            compute_shift_list(frame_list, algorithm='asterism-matching',
                               skip_failure=False)

    @pytest.mark.parametrize('cval,expct_cval', [(np.nan, np.nan),
                                                 ('median', 1),
                                                 ('mean', 1),
                                                 (0, 0)])
    def test_register_framedata_list_skip_failure_true(self, cval, expct_cval):
        frame_list = self.gen_frame_list((512, 1024))
        frame_list[2].data = np.ones((1024, 512))

        reg_list = register_framedata_list(frame_list, clip_output=True,
                                           inplace=False,
                                           algorithm='asterism-matching',
                                           max_control_points=30,
                                           detection_threshold=5,
                                           cval=cval,
                                           skip_failure=True)

        assert_equal(len(frame_list), len(reg_list))
        assert_is_none(reg_list[2].meta['astropop registration_shift_x'])
        assert_is_none(reg_list[2].meta['astropop registration_shift_y'])
        assert_is_none(reg_list[2].meta['astropop registration_rot'])
        assert_equal(reg_list[2].meta['astropop registration'], 'failed')
        assert_equal(reg_list[2].data, np.full(reg_list[2].shape, expct_cval))
        assert_true(np.all(reg_list[2].mask))

        for org, reg in zip(frame_list, reg_list):
            assert_equal(reg.meta['astropop trimmed_section'], '6:-3,2:-23')
            # x: 6:-3, y: 2:-23, since frame[2] is not available
            assert_equal(reg.shape, (1024-23-2, 512-6-3))

        shift_list = compute_shift_list(frame_list,
                                        algorithm='asterism-matching',
                                        max_control_points=30,
                                        detection_threshold=5,
                                        skip_failure=True)
        shift_list_expt = np.array(self._shifts)
        shift_list_expt[2][:] = np.nan
        assert_almost_equal(shift_list, shift_list_expt, decimal=1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Handle compatibility between FrameData and other data formats."""

import warnings
import itertools
from os import PathLike
import copy as cp
import numpy as np
from astropy.io import fits
from astropy.io.fits.verify import VerifyWarning
from astropy.wcs import WCS, FITSFixedWarning
from astropy.nddata.ccddata import CCDData
from astropy.nddata import StdDevUncertainty, VarianceUncertainty, \
                           InverseVariance
from astropy import units as u


from ..logger import logger
from ..fits_utils import imhdus
from ..py_utils import broadcast


__all__ = ['imhdus', 'EmptyDataError']


_PCs = set(['PC1_1', 'PC1_2', 'PC2_1', 'PC2_2'])
_CDs = set(['CD1_1', 'CD1_2', 'CD2_1', 'CD2_2'])
_KEEP = set(['JD-OBS', 'MJD-OBS', 'DATE-OBS'])
_HDU_UNCERT = 'UNCERT'
_HDU_MASK = 'MASK'
_UNIT_KEY = 'BUNIT'


def _remove_sip_keys(header, wcs):
    """Remove the SIP keys. (Too complex fixing)."""
    kwd = '{}_{}_{}'
    pol = ['A', 'B', 'AP', 'BP']
    for poly in pol:
        order = wcs.sip.__getattribute__(f'{poly.lower()}_order')
        for i in [f'{poly}_ORDER', f'{poly}_DMAX']:
            if i in header.keys():
                header.remove(i)
        for i, j in itertools.product(range(order), repeat=2):
            key = kwd.format(poly, i, j)
            if key in header.keys():
                header.remove(key)
    return header


def extract_header_wcs(header):
    """Get a header (or dict) and extract a WCS based on the keys.

    Parameters
    ----------
    header: `~astropy.fits.Header` or dict_like
        Header to extract the WCS.
    logger: `~logging.Logger` (optional)
        Logger instance to log warnings.

    Returns
    -------
    header: `~astropy.fits.Header`
        Header cleaned from WCS keys.
    wcs: `~astropy.wcs.WCS` os `None`
        World Coordinate Sistem extracted from the header.
    """
    header = header.copy()

    # First, check if there is a WCS. If not, return header and None WCS
    try:
        with warnings.catch_warnings():
            # silent WCS anoying warnings
            warnings.filterwarnings("ignore", category=FITSFixedWarning)
            wcs = WCS(header, relax=True, fix=True)
            if not wcs.wcs.ctype[0]:
                wcs = None
    except Exception as e:
        logger.warning('An error raised when extracting WCS: %s', e)
        wcs = None

    if wcs is not None:
        # Delete wcs keys, except time ones
        for k in wcs.to_header(relax=True).keys():
            if k not in _KEEP and k in header.keys():
                header.remove(k)

        # Astropy uses PC. So remove CD if any
        # The reverse case should not be allowed by astropy
        if (_PCs & set(wcs.to_header(relax=True))) and \
           (_CDs & set(header.keys())):
            for k in _CDs:
                if k in header.keys():
                    header.remove(k)

        # Check and remove remaining SIP coefficients
        if wcs.sip is not None:
            header = _remove_sip_keys(header, wcs)

    return (header, wcs)


def _merge_and_clean_header(meta, header, wcs):
    """Merge meta and header and clean the WCS and spurious keys."""
    for i in (meta, header, wcs):
        if not isinstance(i, (dict, fits.Header, WCS)) and i is not None:
            raise TypeError(f'{i} is not a compatible format.')

    if header is not None:
        header = fits.Header(header)
        header.strip()
    else:
        header = fits.Header()

    if meta is not None:
        meta = fits.Header(meta)
        meta.strip()
    else:
        meta = fits.Header()

    meta.update(header)

    # strip blank cards
    meta.remove('', ignore_missing=True)

    # extract history and comments from meta
    if 'history' in meta:
        history = meta['history']
        history = list(broadcast(history).iters[0])
        del meta['history']
    else:
        history = []

    if 'comment' in meta:
        comment = meta['comment']
        comment = list(broadcast(comment).iters[0])
        del meta['comment']
    else:
        comment = []

    # extract wcs from header
    meta, wcs_ = extract_header_wcs(meta)
    wcs = wcs_ if wcs is None else wcs
    return meta, wcs, history, comment


class EmptyDataError(ValueError):
    """Operation not handled by empty MemMapArray containers."""


def _extract_fits(obj, hdu=0, unit=None, hdu_uncertainty=_HDU_UNCERT,
                  hdu_mask=_HDU_MASK, unit_key=_UNIT_KEY):
    """Extract data and meta from FITS files and HDUs."""
    if isinstance(obj, (str, bytes, PathLike)):
        hdul = fits.open(obj)
    elif isinstance(obj, imhdus):
        hdul = fits.HDUList([obj])
    elif isinstance(obj, fits.HDUList):
        hdul = obj
    else:
        raise TypeError(f'Object {obj} not recognized.')

    # from now, assumes obj is HDUList
    # Check if the data HDU contain data
    if hdul[hdu].data is None and isinstance(hdul[hdu], imhdus):
        if hdu == 0:
            hdu_data = None
        else:
            raise ValueError('Informed data HDU do not contain any data.')
    else:
        hdu_data = hdu

    # scan hdus for the first with image data if the first hdu don't
    for ind in range(len(hdul)):
        hdu_i = hdul[ind]
        if not isinstance(hdu_i, imhdus) or hdu_i.data is None:
            continue
        hdu_data = ind
        if hdu_data != hdu:
            logger.info('First hdu with image data: %i', ind)
        break

    hdu_data = hdul[hdu_data]
    res = {}
    res['data'] = hdu_data.data
    res['meta'] = hdu_data.header

    unit_h = res['meta'].get(unit_key, None)
    if unit_h is None or unit_h == '':
        res['unit'] = unit
    else:
        try:
            unit_h = u.format.Fits.parse(unit_h)
        except ValueError:
            unit_h = u.Unit(unit_h)
        if unit_h != unit and unit is not None:
            raise ValueError('unit and unit_key got incompatible results.')
        res['unit'] = unit_h

    if hdu_uncertainty in hdul:
        hdu_unct = hdul[hdu_uncertainty]
        logger.debug('Uncertainty hdu found. Assuming standard uncertainty.')
        res['uncertainty'] = hdu_unct.data

    if hdu_mask in hdul:
        hdu_mask = hdul[hdu_mask]
        res['mask'] = hdu_mask.data

    if isinstance(obj, (str, bytes, PathLike)):
        hdul.close()

    return res


def _extract_ccddata(ccd):
    """Extract data and meta from CCDData objects."""
    res = {}
    for i in ('data', 'unit', 'meta', 'mask'):
        res[i] = cp.copy(getattr(ccd, i))

    uunit = None
    uncert = ccd.uncertainty
    if uncert is None:
        uncert = None
    elif isinstance(uncert, StdDevUncertainty):
        uunit = uncert.unit
        uncert = uncert.array
    elif isinstance(uncert, VarianceUncertainty):
        uunit = uncert.unit**0.5
        uncert = np.sqrt(uncert.array)
    elif isinstance(uncert, InverseVariance):
        uunit = uncert.unit**(-0.5)
        uncert = np.sqrt(1/uncert.array)
    else:
        raise TypeError(f'Unrecognized uncertainty type {uncert.__class__}.')

    if uunit != res['unit'] and uunit is not None:
        raise ValueError(f'Uncertainty with unit {uunit} different from the '
                         f'data {res["unit"]}')
    res['uncertainty'] = uncert

    return res


def _to_ccddata(frame):
    """Translate a FrameData to a CCDData."""
    data = np.array(frame.data)
    unit = frame.unit
    meta = dict(frame.header)
    wcs = frame.wcs
    uncertainty = frame._unct
    if uncertainty.empty:
        uncertainty = None
    else:
        uncertainty = StdDevUncertainty(uncertainty, unit=unit)
    mask = np.array(frame.mask)

    return CCDData(data, unit=unit, meta=meta, wcs=wcs,
                   uncertainty=uncertainty, mask=mask)


def _to_hdu(frame, hdu_uncertainty=_HDU_UNCERT, hdu_mask=_HDU_MASK,
            unit_key=_UNIT_KEY, wcs_relax=True, **kwargs):
    """Translate a FrameData to an HDUList."""
    data = frame.data.copy()

    # Clean header
    header = fits.Header(frame.header)
    no_fits_std = kwargs.pop('no_fits_standard_units', False)

    if frame.wcs is not None:
        header.extend(frame.wcs.to_header(relax=wcs_relax),
                      update=True)

    if no_fits_std:
        unit_string = frame.unit.to_string()
    else:
        unit_string = u.format.Fits.to_string(frame.unit)

    header[unit_key] = unit_string

    for i in frame.history:
        header['history'] = i
    hdul = fits.HDUList(fits.PrimaryHDU(data, header=header))

    if hdu_uncertainty and not frame._unct.empty:
        uncert = frame.uncertainty
        uncert_h = fits.Header()
        uncert_h[unit_key] = unit_string
        hdul.append(fits.ImageHDU(uncert, header=uncert_h,
                                  name=hdu_uncertainty))

    if hdu_mask is not None and not frame._mask.empty:
        mask = frame.mask
        if np.issubdtype(mask.dtype, np.bool):
            # Fits do not support bool
            mask = mask.astype('uint8')
        hdul.append(fits.ImageHDU(mask, name=hdu_mask))

    return hdul


def _write_fits(frame, filename, overwrite=True, **kwargs):
    """Write a framedata to a fits file."""
    # FIXME: electron unit is not compatible with fits standards
    with warnings.catch_warnings():
        # silent hierarch warnings
        warnings.filterwarnings("ignore", category=VerifyWarning)
        frame.to_hdu(**kwargs).writeto(filename, overwrite=overwrite,
                                       output_verify='silentfix')



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .aperture import aperture_photometry
from .detection import (background, sepfind, daofind, calc_fwhm,
                        recenter_sources, starfind, gen_filter_kernel,
                        sources_mask)
from .solve_photometry import (solve_photometry_median,
                               solve_photometry_average,
                               solve_photometry_montecarlo)

psf_available_models = ['gaussian', 'moffat']
photometry_available_methods = ['aperture']
solve_photometry_available_methods = ['median', 'average', 'montecarlo']

__all__ = ['aperture_photometry', 'background', 'sepfind', 'daofind',
           'calc_fwhm', 'recenter_sources', 'starfind',
           'solve_photometry_median', 'solve_photometry_average',
           'solve_photometry_montecarlo', 'gen_filter_kernel',
           'sources_mask']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Module to handle local (offline) astronomical catalogs."""

import abc
import numpy as np
from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy import units as u

from ..astrometry.coords_utils import guess_coordinates
from .base_catalog import _BasePhotometryCatalog, match_indexes


__all__ = ['TableCatalog', 'ASCIICatalog', 'FITSCatalog']


class _LocalCatalog(_BasePhotometryCatalog, abc.ABC):
    type = 'local'
    id_key = None
    ra_key = None
    dec_key = None
    flux_key = None
    flux_error_key = None
    flux_unit = None
    available_filters = None
    prepend_id_key = False
    bibcode = None

    _table = None  # Where the data is stored.

    @property
    def skycoords(self):
        if self.ra_key is not None and self.dec_key is not None:
            tabs = guess_coordinates(self._table[self.ra_key],
                                     self._table[self.dec_key],
                                     skycoord=False)
            return SkyCoord(tabs[0], tabs[1], unit=('degree', 'degree'))
        return None

    @property
    def id(self):
        if self.id_key is None:
            return None
        return np.array(self._table[self.id_key])

    def _query_index(self, center=None, radius=None):
        """Return the index of the points within a radius distance of center
        point.

        If center is None, return all
        """
        if center is None:
            return np.arange(0, len(self._table), 1)

        center = self._get_center(center)
        radius = self._get_radius(radius)
        coords = self.skycoords
        sep = coords.separation(center)
        filt = np.where(sep <= radius*u.degree)
        return filt

    def query_ra_dec(self, center=None, radius=None):
        """Query coordinates in a region of the catalog."""
        filt = self._query_index(center, radius)
        coords = self.skycoords
        return coords.ra.degree[filt], coords.dec.degree[filt]

    def query_flux(self, center=None, radius=None):
        """Query the flux data in a region of the catalog."""
        filt = self._query_index(center, radius)
        if self.flux_key is not None:
            flux = np.array(self._table[self.flux_key][filt])
        else:
            flux = np.zeros(len(filt))
            flux.fill(np.nan)
        if self.flux_error_key is not None:
            error = np.array(self._table[self.flux_error_key][filt])
        else:
            error = np.zeros(len(filt))
            error.fill(np.nan)
        return flux, error

    def query_id(self, center=None, radius=None):
        """Query coordinates in a region of the catalog."""
        filt = self._query_index(center, radius)
        return self.id[filt]

    def match_objects(self, ra, dec, limit_angle='2 arcsec'):
        """Query the informations in the catalog from a list of ra and dec
        coordinates, matching the stars by a limit_angle."""
        rac, decc = self.query_ra_dec()

        indx = match_indexes(ra, dec, rac, decc, limit_angle)

        nstars = len(ra)
        m_id = self.query_id()
        m_id = np.array([m_id[i] if i != -1 else '' for i in indx])
        m_ra, m_dec = self.query_ra_dec()
        m_ra = np.array([m_ra[i] if i != -1 else np.nan for i in indx])
        m_dec = np.array([m_dec[i] if i != -1 else np.nan for i in indx])
        try:
            flux, error = self.query_flux()
            m_f = np.array([flux[i] if i != -1 else np.nan
                            for i in indx])
            m_e = np.array([error[i] if i != -1 else np.nan
                            for i in indx])
        except NotImplementedError:
            m_f = np.zeros(nstars)
            m_f.fill(np.nan)
            m_e = np.zeros(nstars)
            m_e.fill(np.nan)

        return np.array(list(zip(m_id, m_ra, m_dec, m_f, m_e)),
                        dtype=np.dtype([('id', m_id.dtype),
                                        ('ra', m_ra.dtype),
                                        ('dec', m_dec.dtype),
                                        ('flux', m_f.dtype),
                                        ('flux_error', m_e.dtype)]))

    def match_object_ids(self, ra, dec, limit_angle='2 arcsec'):
        return self.match_objects(ra, dec, limit_angle=limit_angle)['id']


class TableCatalog(_LocalCatalog):
    """Local catalog loaded from anything that can be converted directly to
    an `~astropy.table.Table`."""
    type = 'local'

    def __init__(self, table, id_key=None, ra_key=None, dec_key=None,
                 flux_key=None, flux_error_key=None, flux_unit=None,
                 available_filters=None, prepend_id_key=False, bibcode=None):
        self._table = Table(table)

        self.id_key = id_key
        self.ra_key = ra_key
        self.dec_key = dec_key
        self.flux_key = flux_key
        self.flux_error_key = flux_error_key
        self.flux_unit = flux_unit
        self.available_filters = available_filters
        self.prepend_id_key = prepend_id_key
        self.bibcode = bibcode


class ASCIICatalog(TableCatalog):
    """Local catalog read from a ASCII table file."""
    def __init__(self, filename, id_key=None, ra_key=None, dec_key=None,
                 flux_key=None, flux_error_key=None, flux_unit=None,
                 available_filters=None, prepend_id_key=False, bibcode=None,
                 **reader_kwargs):
        """
        **reader_kwargs : kwargs to be passed to the Table.read function
        """
        table = Table.read(filename, **reader_kwargs)

        super().__init__(table, id_key, ra_key, dec_key, flux_key,
                         flux_error_key, flux_unit, available_filters,
                         prepend_id_key, bibcode)


class FITSCatalog(ASCIICatalog):
    """Local catalog read from a FITS table."""
